import numpy as np
import pandas as pd
from pkg_resources import resource_filename


def load_sample_data(timevary):
    '''Load data that is part of the zepid package. This data set comes
    from simulated data from Jessie Edwards (thanks Jess!). This data
    is used for all examples on zepid.readthedocs.io and part of my
    other repository Python-for-Epidemiologists

    timevary:
        True    -produces a repeated follow-up data set
        False   -produces a single observation per subject

    Variables:
    -id: participant unique ID (multiple observations per person)
    -enter: start of time period
    -out: end of time period
    -male: indicator variable for male (1 = yes)
    -age0: baseline age (at enter = 0)
    -cd40: baseline CD4 T cell count (at enter = 0)
    -dvl0: baseline viral load data (at enter = 0)
    -cd4: CD4 T cell count at that follow-up visit
    -dvl: viral load at that follow-up visit
    -art: indicator of whether ART (antiretroviral treatment) was
          prescribed at that time
    -drop: indicator of whether individual dropped out of the study (1 = yes)
    -dead: indicator for death at end of follow-up period (1 = yes)
    -t: total time contributed
    '''
    cols = ['id', 'enter', 'out', 'male', 'age0', 'cd40', 'dvl0', 'cd4', 'dvl', 'art', 'drop', 'dead']
    df = pd.read_csv(resource_filename('zepid', 'datasets/data.txt'),
                     delim_whitespace=True, header=None, names=cols, index_col=False)
    df.sort_values(by=['id', 'enter'], inplace=True)
    if timevary is True:
        return df
    else:
        dfi = df.loc[df.groupby('id').cumcount()==0][['id', 'male', 'age0', 'cd40', 'cd4', 'dvl0', 'art']].copy()
        dfo = df.loc[df.id != df.id.shift(-1)][['id', 'dead', 'drop', 'out']].copy()
        dfo.loc[dfo['drop'] == 1, 'dead'] = np.nan
        dff = pd.merge(dfi, dfo, left_on='id', right_on='id')
        dff.rename(columns={'out': 't'}, inplace=True)
        dff.drop('drop', axis=1, inplace=True)
        return dff





from .utils import (risk_ci, ir_ci, rr, rd, nnt, oddsratio, irr, ird,
                    acr, paf, prop_to_odds, odds_to_prop,
                    counternull_pvalue, semibayes,
                    ppv_converter, npv_converter, screening_cost_analyzer)


import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.genmod.families import links
from zepid.causal.ipw import propensity_score


class AIPW:
    """Implementation of augmented inverse probablity weight estimator as described in Funk et al. American Journal
    of Epidemiology 2011;173(7):761-767. The exact formulas used are available in Table 1 and Equation 1. Properties
    of the doubly robust estimator are often misunderstood, so we direct the reader to  Keil et al. AJE
    2018;187(4):891-892. SimplyDoubleRobust only supports a binary exposure and binary outcome. Also this model does
    not deal with missing data or time varying exposures

    df:
        -pandas DataFrame object containing all variables of interest
    exposure:
        -column name of the exposure variable. Currently only binary is supported
    outcome:
        -column name of the outcome variable. Currently only binary is supported
    """

    def __init__(self, df, exposure, outcome):
        self.df = df.copy()
        self._exposure = exposure
        self._outcome = outcome
        self._fit_exposure_model = False
        self._fit_outcome_model = False
        self._generated_ci = False
        self.riskdiff = None
        self.riskratio = None
        self._exp_model = None
        self._out_model = None

    def exposure_model(self, model, print_model_results=True):
        """Used to specify the propensity score model. Model used to predict the exposure via a logistic regression
        model

        model:
            -Independent variables to predict the exposure. Example) 'var1 + var2 + var3'
        print_model_results:
            -Whether to print the fitted model results. Default is True (prints results)
        """
        self._exp_model = self._exposure + ' ~ ' + model
        fitmodel = propensity_score(self.df, self._exp_model, mresult=print_model_results)
        self.df['ps'] = fitmodel.predict(self.df)
        self._fit_exposure_model = True

    def outcome_model(self, model, print_model_results=True):
        """Used to specify the outcome model. Model used to predict the outcome via a logistic regression model

        model:
            -Independent variables to predict the outcome. Example) 'var1 + var2 + var3 + var4'
        print_model_results:
            -Whether to print the fitted model results. Default is True (prints results)
        """

        self._out_model = self._outcome + ' ~ ' + model
        f = sm.families.family.Binomial(sm.families.links.logit)
        log = smf.glm(self._out_model, self.df, family=f).fit()
        if print_model_results:
            print('\n----------------------------------------------------------------')
            print('MODEL: ' + self._out_model)
            print('-----------------------------------------------------------------')
            print(log.summary())

        dfx = self.df.copy()
        dfx[self._exposure] = 1
        self.df['pY1'] = log.predict(dfx)
        dfx = self.df.copy()
        dfx[self._exposure] = 0
        self.df['pY0'] = log.predict(dfx)
        self._fit_outcome_model = True

    def fit(self):
        """Once the exposure and outcome models are specified, we can estimate the Risk Ratio and Risk Difference.
        This function generates the estimated risk difference and risk ratio.  To view results, use AIPW.summary() For
        confidence intervals, a bootstrap procedure should be used
        """
        if ((self._fit_exposure_model == False) or (self._fit_exposure_model == False)):
            raise ValueError('The exposure and outcome models must be specified before the doubly robust estimate can '
                             'be generated')

        # Doubly robust estimator for exposed
        self.df['dr1'] = np.where(self.df[self._exposure] == 1,
                                  ((self.df[self._outcome]) / self.df['ps']) - (((self.df['pY1'] * (1 - self.df['ps']))
                                                                                 / (self.df['ps']))),
                                  self.df['pY1'])

        # Doubly robust estimator for unexposed
        self.df['dr0'] = np.where(self.df[self._exposure] == 0,
                                  (self.df['pY0']),
                                  ((self.df[self._outcome]) / (1 - self.df['ps']) - (
                                              ((self.df['pY0']) * (self.df['ps']))
                                              / (1 - self.df['ps']))))

        # Generating estimates for the risk difference and risk ratio
        self.riskdiff = np.mean(self.df['dr1']) - np.mean(self.df['dr0'])
        self.riskratio = np.mean(self.df['dr1']) / np.mean(self.df['dr0'])

    def summary(self, decimal=4):
        """Prints a summary of the results for the doubly robust estimator.

        decimal:
            -number of decimal places to display in the result
        """
        if (self._fit_exposure_model == False) or (self._fit_exposure_model == False):
            raise ValueError('The exposure and outcome models must be specified before the double robust estimate can '
                             'be generated')

        print('----------------------------------------------------------------------')
        print('Risk Difference: ', round(float(self.riskdiff), decimal))
        print('Risk Ratio: ', round(float(self.riskratio), decimal))
        print('----------------------------------------------------------------------')



import warnings
import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.genmod.families import links


class TimeFixedGFormula:
    '''Time-fixed implementation of the g-formula, also referred to as the g-computation
    algorithm formula. This implementation has three options for the treatment courses:
    
    Key options for treatments
        all     -all individuals are given treatment
        none    -no individuals are given treatment
    Custom treatment
                -create a custom treatment. When specifying this, the dataframe must be 
                 referred to as 'g' The following is an example that selects those whose
                 age is 30 or younger and are females
                 ex) treatment="((g['age0']<=30) & (g['male']==0))
    
    Currently, only supports binary or continuous outcomes. For binary outcomes a logistic regression 
    model to predict probabilities of outcomes via statsmodels. For continuous outcomes a linear regression
    model is used to predict outcomes. 
    Binary and multivariate exposures are supported. For binary exposures, a string object of the column name for 
    the exposure of interest should be provided. For multivariate exposures, a list of string objects corresponding
    to disjoint indicator terms for the exposure should be provided. Multivariate exposures require the user to 
    custom specify treatments when fitting the g-formula. A list of the custom treatment must be provided and be 
    the same length as the number of disjoint indicator columns.
    
    See Snowden et al. (2011) for a good description of the time-fixed g-formula (also freely available on 
    PubMed). See http://zepid.readthedocs.io/en/latest/ for an example (highly recommended)
    
    Inputs:
    df:
        -pandas dataframe containing the variables of interest
    exposure:
        -exposure variable label / column name, or a list of disjoint indicator exposures
    outcome:
        -outcome variable label / column name
    outcome_type:
        -outcome variable type. Currently only 'binary' or 'continuous' variable types are supported
    '''
    def __init__(self, df, exposure, outcome, outcome_type='binary'):
        self.gf = df.copy()
        self.exposure = exposure
        self.outcome = outcome
        if (outcome_type == 'binary') or (outcome_type == 'continuous'):
            self.outcome_type = outcome_type
        else:
            raise ValueError('Only binary or continuous outcomes are currently supported. Please specify "binary" or '
                             '"continuous"')
        self.model_fit = False
    
    def outcome_model(self, model, print_model_results=True):
        '''Build the model for the outcome. This is also referred to at the Q-model. This must be specified
        before the fit function. If it is not, an error will be raised.
        
        Input:
        
        model:
            -variables to include in the model for predicting the outcome. Must be contained within the input
             pandas dataframe when initialized. Model form should contain the exposure. Format is the same as 
             the functional form, i.e. 'var1 + var2 + var3 + var4'
        print_model_results:
            -whether to print the logistic regression results to the terminal. Default is True
        '''
        if self.outcome_type == 'binary':
            linkdist = sm.families.family.Binomial(sm.families.links.logit)
        else:
            linkdist = sm.families.family.Gaussian(sm.families.links.identity)
        
        # Modeling the outcome
        m = smf.glm(self.outcome+' ~ '+model, self.gf, family=linkdist)
        self.outcome_model = m.fit()
        if print_model_results is True:
            # Warning to make users aware that some amount of observations are dropped
            if self.gf.shape[0] != self.outcome_model.nobs:
                warnings.warn('Notice: ' +
                              str(self.gf.shape[0] - self.outcome_model.nobs) +
                              ' observations were dropped due to missing data for at least one of the variables in the '
                              'specified regression model')
            print(self.outcome_model.summary())
        self.model_fit = True

    def fit(self, treatment):
        '''Fit the parametric g-formula as specified. Binary and multivariate treatments are available.
        This implementation has three options for the binary treatment courses:
    
        all     -all individuals are given treatment
        none    -no individuals are given treatment
        custom  -create a custom treatment. When specifying this, the dataframe must be 
                 referred to as 'g' The following is an example that selects those whose
                 age is 25 or older and are females
                 ex) treatment="((g['age0']>=25) & (g['male']==0))
        
        For multivariate treatments, the user must specify custom treatments
        
        To obtain the confidence intervals, use a bootstrap. See online documentation for 
        an example: http://zepid.readthedocs.io/en/latest/
        
        treatment:
            -specified treatment course. Either a string object for binary treatments or a list of custom
             treatments as strings
        '''
        if self.model_fit is False:
            raise ValueError('Before the g-formula can be calculated, the outcome model must be specified')
        if (type(treatment) != str) and (type(treatment) != list):
            raise ValueError('Specified treatment must be a string object or a list of string objects')
        
        # Setting outcome as blank
        g = self.gf.copy()
        
        # Setting treatment (either multivariate or binary)
        if type(self.exposure) == list:  # Multivariate exposure
            if (treatment == 'all') or (treatment == 'none'): # Check to make sure custom treatment
                raise ValueError('A multivariate exposure has been specified. A custom treatment must be '
                                 'specified by the user')
            else:
                if len(self.exposure) != len(treatment): # Check to make sure same about of treatments specified
                    raise ValueError('The list of custom treatment conditions must be the same size as the number of '
                                     'treatments')
                for i in range(len(self.exposure)):
                    g[self.exposure[i]] = np.where(eval(treatment[i]),1,0)
                if np.sum(np.where(g[self.exposure].sum(axis=1)>1,1,0)) > 1:
                    warnings.warn('It looks like your specified treatment strategy results in some individuals '
                                  'receiving at least two exposures. Reconsider how the custom treatments are '
                                  'specified')
        
        else:  # Binary exposure
            if type(treatment) == list:
                raise ValueError('A binary exposure is specified. Treatment plan should be a string object')
            if treatment == 'all':
                g[self.exposure] = 1
            elif treatment == 'none':
                g[self.exposure] = 0
            else:  # custom exposure pattern
                g[self.exposure] = np.where(eval(treatment),1,0)
        
        # Getting predictions
        g[self.outcome] = np.nan
        g[self.outcome] = self.outcome_model.predict(g)
        self.marginal_outcome = np.mean(g[self.outcome])
        self.predicted_df = g



import warnings
import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.genmod.families import links


class TimeVaryGFormula:
    """Time-varying implementation of the g-formula, also referred to as the g-computation
    algorithm formula. This implementation has four options for the treatment courses:
    
    Key options for treatments
        all     -all individuals are given treatment
        none    -no individuals are given treatment
        natural -individuals retain their observed treatment
        custom treatment
                -create a custom treatment. When specifying this, the dataframe must be 
                 referred to as 'g' The following is an example that selects those whose
                 age is 25 or older and are females
                 ex) treatment="((g['age0']>=25) & (g['male']==0))
    
    Currently, only supports a binary exposure and a binary outcome. Uses a logistic regression model to predict
    exposures and outcomes via statsmodels. See Keil et al. (2014) for a good description of the time-varying
    g-formula. See http://zepid.readthedocs.io/en/latest/ for an example (highly recommended)
    
    IMPORTANT CONSIDERATIONS:
    1) TimeVaryGFormula increases by time unit increase of one. Your input dataset should reflect this
    2) Only binary exposures and binary outcomes are supported
    3) Binary and continuous covariates are supported
    4) The labeling of the covariate models is important. They are fit in the order that they are labeled!
    5) Fit the natural course model first and compare to the observed data. They should be similar
    6) Check to make sure the predicted values are reasonable. If not, you may need to code in restrictions into
        'restriction', 'recode', or 'in_recode' statements in the fitted models

    Process for the g-formula monte carlo:
        1) run lines in "in_recode"
        2) time-varying covariates, order ascending in from "labels"
            a) predict time-varying covariate
            b) run lines in "recode" from "add_covariate_model()"
        3) predict exposure / apply exposure pattern
        4) predict outcome
        5) run lines in "out_recode"
        6) lag variables in "lags"
        7) append current-time rows to full dataframe
        8) Repeat till t_max is met

    Inputs:
    df:
        -pandas dataframe containing the variables of interest
    exposure:
        -exposure variable label / column name
    outcome:
        -outcome variable label / column name
    time:
        -time variable label / column name
    """

    def __init__(self, df, idvar, exposure, outcome, time_in, time_out):
        self.gf = df.copy()
        self.idvar = idvar
        self.exposure = exposure
        self.outcome = outcome
        self.exp_model = None
        self.out_model = None
        self.time_in = time_in
        self.time_out = time_out
        self._exposure_model_fit = False
        self._outcome_model_fit = False
        self._covariate_models = []
        self._covariate_model_index = []
        self._covariate = []
        self._covariate_type = []
        self._covariate_recode = []
        self.predicted_outcomes = None

    def exposure_model(self, model, restriction=None, print_results=True):
        """Build the model for the exposure. This must be specified before the fit function. If it is not,
        an error will be raised.
        
        model:
            -variables to include in the model for predicting the outcome. Must be contained within the input
             pandas dataframe when initialized. Format is the same as the functional form
             Example) 'var1 + var2 + var3 + var4'
        restriction:
            -used to restrict the population to fit the logistic regression model to. Useful for Intent-to-Treat
             model fitting. The pandas dataframe must be referred to as 'g'
             Example) "g['art']==1"
        print_results:
            -whether to print the logistic regression results to the terminal. Default is True
        """
        g = self.gf.copy()
        if restriction is not None:
            g = g.loc[eval(restriction)].copy()
        linkdist = sm.families.family.Binomial(sm.families.links.logit)
        self.exp_model = smf.glm(self.exposure + ' ~ ' + model, g, family=linkdist).fit()
        if print_results:
            print(self.exp_model.summary())
        self._exposure_model_fit = True

    def outcome_model(self, model, restriction=None, print_results=True):
        """Build the model for the outcome. This must be specified before the fit function. If it is not,
        an error will be raised.
        
        Input:
        
        model:
            -variables to include in the model for predicting the outcome. Must be contained within the input
             pandas dataframe when initialized. Format is the same as the functional form,
             i.e. 'var1 + var2 + var3 + var4'
        restriction:
            -used to restrict the population to fit the logistic regression model to. Useful for Intent-to-Treat
             model fitting. The pandas dataframe must be referred to as 'g'
             Example) "g['art']==1"
        print_results:
            -whether to print the logistic regression results to the terminal. Default is True
        """
        g = self.gf.copy()
        if restriction is not None:
            g = g.loc[eval(restriction)].copy()
        linkdist = sm.families.family.Binomial(sm.families.links.logit)
        self.out_model = smf.glm(self.outcome + ' ~ ' + model, g, family=linkdist).fit()
        if print_results:
            print(self.out_model.summary())
        self._outcome_model_fit = True

    def add_covariate_model(self, label, covariate, model, restriction=None, recode=None, var_type='binary',
                            print_results=True):
        """
        Build the model for the specified covariate. This is to deal with time-varying confounders.
        Does NOT have to be specified, unlike the exposure and outcome models. The order in which these
        models are fit is based on the provided integer labels
        
        Input:
        
        label:
            -integer label for the covariate model. Covariate models are fit in ascending order within 
             TimeVaryGFormula
        covariate:
            -variable to be predicted
        model:
            -variables to include in the model for predicting the outcome. Must be contained within the input
             pandas dataframe when initialized. Format is the same as the functional form,
             i.e. 'var1 + var2 + var3 + var4'
        restriction:
            -used to restrict the population to fit the logistic regression model to. Useful for Intent-to-Treat
             model fitting. The pandas dataframe must be referred to as 'g'
             Example) "g['art']==1"
        recode:
            -This variable is vitally important for various functional forms implemented later in models. This
             is used to run some background code to recreate functional forms as the g-formula is fit via fit()
             For an example, let's say we have age but we want the functional form to be cubic. For this, we 
             would set the recode="g['']" Similar to TimeFixedGFormula, 'g' must be specified as the data frame 
             object with the corresponding indexes. Also lines of executable code should end with ';', so Python
             knows that the line ends there. My apologies for this poor solution... I am working on a better way
        var_type:
            -type of variable that the covariate is. Current options include 'binary' or 'continuous'
        print_results:
            -whether to print the logistic regression results to the terminal. Default is True
        """
        if type(label) is not int:
            raise ValueError('Label must be an integer')

        # Building predictive model
        g = self.gf.copy()
        if restriction is not None:
            g = g.loc[eval(restriction)].copy()
        if var_type == 'binary':
            linkdist = sm.families.family.Binomial(sm.families.links.logit)
            m = smf.glm(covariate + ' ~ ' + model, g, family=linkdist)
        elif var_type == 'continuous':
            m = smf.gls(covariate + ' ~ ' + model, g)
        else:
            raise ValueError('Only binary or continuous covariates are currently supported')

        f = m.fit()
        if print_results:
            print(f.summary())

        # Adding to lists, it is used to predict variables later on for the time-varying...
        self._covariate_models.append(f)
        self._covariate_model_index.append(label)
        self._covariate.append(covariate)
        self._covariate_type.append(var_type)
        if recode is None:
            self._covariate_recode.append('None')  # Must be string for exec() to use later
        else:
            self._covariate_recode.append(recode)

    def fit(self, treatment, lags=None, sample=10000, t_max=None, in_recode=None, out_recode=None):
        """
        Implementation of fit
        
        treatment:
            -treatment strategy
        lags:
            -variables to generate a lagged variable for. It should be a dictionary with the variable name
             as the key and the lagged variable name as the value.
             Ex) {'art':'lagged_art'}
        sample:
            -number of individuals to sample from the original data with replacement. It should be a large 
             number. Default is 10000
        t_max:
            -maximum time to run g-formula until. Default is None, which uses the maximum time of the input 
             dataframe. Input should be integer (but will also accept floats)
        in_recode:
            -on the fly recoding of variables done before the loop starts. Needed to do any kind of functional
             forms for entry times
        """
        if self._exposure_model_fit is False:
            raise ValueError('Before the g-formula can be calculated, the exposure model must be specified')
        if self._outcome_model_fit is False:
            raise ValueError('Before the g-formula can be calculated, the outcome model must be specified')
        if (type(treatment) != str) and (type(treatment) != list):
            raise ValueError('Specified treatment must be a string object')

        # Getting data all set
        gs = self.gf.loc[(self.gf.groupby(self.idvar).cumcount() == 0) == True].sample(n=sample, replace=True)
        gs['uid_g_zepid'] = [v for v in range(sample)]

        # Background preparations
        gs[self.outcome] = 0

        # getting maximum time steps to run g-formula
        if t_max is None:
            t_max = np.max(self.gf[self.time_out])

        # fitting g-formula piece by piece
        gs['continuous_standerror'] = np.random.normal(size=gs.shape[0])
        g = gs.copy()
        for i in range(int(t_max)):
            g = g.loc[g[self.outcome] == 0].copy()
            g[self.time_in] = i
            if in_recode is not None:
                exec(in_recode)

            # predict time-varying covariates
            if len(self._covariate_models) != 0:
                for j in (sorted(range(len(self._covariate_model_index)),
                                 key=self._covariate_model_index.__getitem__)):
                    g[self._covariate[j] + 'pred'] = self._covariate_models[j].predict(g)
                    if self._covariate_type[j] == 'binary':
                        g[self._covariate[j]] = np.random.binomial(1, g[self._covariate[j] + 'pred'], size=g.shape[0])
                    if self._covariate_type[j] == 'continuous':
                        g[self._covariate[j]] = np.random.normal(loc=g[self._covariate[j] + 'pred'],
                                                                 scale=np.std(self._covariate_models[j].resid),
                                                                 size=g.shape[0])
                    exec(self._covariate_recode[j])

            # predict exposure when customized treatments
            if treatment == 'all':
                g[self.exposure] = 1
            elif treatment == 'none':
                g[self.exposure] = 0
            elif treatment == 'natural':
                g[self.exposure + 'pred'] = self.exp_model.predict(g)
                g[self.exposure] = np.random.binomial(1, g[self.exposure + 'pred'], size=g.shape[0])
            else:  # custom exposure pattern
                g[self.exposure + 'pred'] = self.exp_model.predict(g)
                g[self.exposure] = np.random.binomial(1, g[self.exposure + 'pred'], size=g.shape[0])
                g[self.exposure] = np.where(eval(treatment), 1, 0)

            # predict outcome
            g['pevent_zepid'] = self.out_model.predict(g)
            g[self.outcome] = np.random.binomial(1, g['pevent_zepid'], size=g.shape[0])
            g[self.time_out] = i + 1

            # executing any counting code or else before appending
            if out_recode is not None:
                exec(out_recode)

            # recreating lagged variables
            if lags is not None:
                g.drop(list(lags.values()), axis=1, inplace=True)
                g.rename(columns=lags, inplace=True)

            # APPEND SIMULATED DATA
            gs = gs.append(g, ignore_index=True)  # sort=False for 0.23.0+

        # so the append loop adds an extra, so we need to drop that extra
        gs.dropna(subset=['pevent_zepid'], inplace=True)
        # Gains new attributes
        self.predicted_outcomes = gs[['uid_g_zepid', self.exposure, self.outcome, self.time_in, self.time_out] +
                                     self._covariate].sort_values(by=['uid_g_zepid',
                                                                      self.time_in]).reset_index(drop=True)



'''Implementations of the parametric g-computation algorithm model, also known as the g-formula. 
Currently, only parametric g-formula is available

TimeFixedGFormula:
    -time-fixed exposure implementation of the g-formula. See Snowden et al. 2011 for an introduction
TimeVaryGFormula:
    -under production...
'''

from .TimeFixed import TimeFixedGFormula
from .TimeVary import TimeVaryGFormula





import warnings
import math
import numpy as np
from scipy import stats
import matplotlib.pyplot as plt
from .utils import propensity_score


class IPTW:
    """
    Calculates the weight for inverse probability of treatment weights through logistic regression.
    Both stabilized or unstabilized weights are implemented. Default is just to calculate the prevalence
    of the treatment in the population.

    df:
        -pandas dataframe object containing all variables of interest
    treatment:
        -Variable name of treatment variable of interest. Must be coded as binary.
         1 should indicate treatment, while 0 indicates no treatment
    stabilized:
        -Whether to return stabilized or unstabilized weights. Input is True/False. Default is
         stabilized weights (True)
    standardize:
        -who to standardize the estimate to. Options are the entire population, the exposed, or
         the unexposed. See Sato & Matsuyama Epidemiology (2003) for details on weighting to exposed/unexposed
         Options for standardization are:
            'population'    :   weight to entire population
            'exposed'       :   weight to exposed individuals
            'unexposed'     :   weight to unexposed individuals
    """
    def __init__(self, df, treatment, stabilized=True, standardize='population'):
        self.df = df.copy()
        self.ex = treatment
        self.stabilized = stabilized
        if standardize in ['population', 'exposed', 'unexposed']:
            self.standardize = standardize
        else:
            raise ValueError('Please specify one of the currently supported weighting schemes: ' +
                             'population, exposed, unexposed')

    def regression_models(self, model_denominator, model_numerator='1', print_model_results=True):
        """
        Logistic regression model(s) for propensity score models. The model denominator must be specified for both
        stabilized and unstabilized weights. The optional argument 'model_numerator' allows specification of the
        stabilization factor for the weight numerator. By default model results are returned

        model_denominator:
            -statsmodels glm format for modeling data. Only includes predictor variables
             Example) 'var1 + var2 + var3'
        model_numerator:
            - statsmodels glm format for modeling data. Only includes predictor variables for the numerator.
             Default ('1') calculates the overall probability. In general this is recommended. If confounding
             variables are included in the numerator, they would later need to be adjusted for.
             Example) 'var1'
        print_model_results:
            -whether to print the model results from the regression models. Default is True
        """
        self.denominator_model = propensity_score(self.df, self.ex + ' ~ ' + model_denominator,
                                                  mresult=print_model_results)
        if self.stabilized is True:
            self.numerator_model = propensity_score(self.df, self.ex + ' ~ ' + model_numerator,
                                                    mresult=print_model_results)
        else:
            if model_numerator != '1':
                warnings.warn('Argument for model_numerator is only used for stabilized=True')

    def fit(self):
        """
        Uses the specified regression models from 'regression_models' to generate the corresponding
        inverse probability of treatment weights

        IPTW will have the Weight attribute which contains a pandas Series of the calculated IPTW
        """
        self.df['propscore'] = self.denominator_model.predict(self.df)
        if self.stabilized:
            n = self.numerator_model.predict(self.df)
        else:
            n = 1
        self.df['denom'] = self.df['propscore']
        self.df['numer'] = n
        self.df['iptw'] = self._weight_calculator(self.df, denominator='denom', numerator='numer')
        self.Weight = self.df['iptw']
        self.ProbabilityDenominator = self.df['denom']
        self.ProbabilityNumerator = self.df['numer']

    def plot_kde(self, bw_method='scott', fill=True, color_e='b', color_u='r'):
        """
        Generates a density plot that can be used to check whether positivity may be violated qualitatively. Note
        input probability variable, not the weight! The kernel density used is SciPy's Gaussian kernel. Either Scott's
        Rule or Silverman's Rule can be implemented.

        bw_method:
            -method used to estimate the bandwidth. Following SciPy, either 'scott' or 'silverman' are
             valid options
        fill:
            -whether to color the area under the density curves. Default is true
        color_e:
            -color of the line/area for the treated group. Default is Blue
        color_u:
            -color of the line/area for the treated group. Default is Red

        Returns matplotlib axes
        """
        x = np.linspace(0, 1, 10000)
        density_t = stats.kde.gaussian_kde(self.df.loc[self.df[self.ex] == 1]['propscore'].dropna(),
                                           bw_method=bw_method)
        density_u = stats.kde.gaussian_kde(self.df.loc[self.df[self.ex] == 0]['propscore'].dropna(),
                                           bw_method=bw_method)
        ax = plt.gca()
        if fill:
            ax.fill_between(x, density_t(x), color=color_e, alpha=0.2, label=None)
            ax.fill_between(x, density_u(x), color=color_u, alpha=0.2, label=None)
        ax.plot(x, density_t(x), color=color_e, label='Treat = 1')
        ax.plot(x, density_u(x), color=color_u, label='Treat = 0')
        ax.set_xlabel('Probability')
        ax.set_ylabel('Density')
        ax.legend()
        return ax

    def plot_boxplot(self):
        """
        Generates a stratified boxplot that can be used to visually check whether positivity may be violated,
        qualitatively.

        Returns matplotlib axes
        """
        boxes = (self.df.loc[self.df[self.ex] == 1]['propscore'].dropna(),
                 self.df.loc[self.df[self.ex] == 0]['propscore'].dropna())
        labs = ['Treat = 1', 'Treat = 0']
        meanpointprops = dict(marker='D', markeredgecolor='black', markerfacecolor='black')
        ax = plt.gca()
        ax.boxplot(boxes, labels=labs, meanprops=meanpointprops, showmeans=True)
        ax.set_ylabel('Probability')
        return ax

    def positivity(self, decimal=3):
        '''Use this to assess whether positivity is a valid assumption. Note that this should only be used for
        stabilized weights generated from IPTW. This diagnostic method is based on recommendations from
        Cole SR & Hernan MA (2008). For more information, see the following paper:
        Cole SR, Hernan MA. Constructing inverse probability weights for marginal structural models.
        American Journal of Epidemiology 2008; 168(6):656–664.

        decimal:
            -number of decimal places to display. Default is three
        '''
        if not self.stabilized:
            warnings.warn('Positivity should only be used for stabilized IPTW')
        avg = float(np.mean(self.df['iptw'].dropna()))
        mx = np.max(self.df['iptw'].dropna())
        mn = np.min(self.df['iptw'].dropna())
        sd = float(np.std(self.df['iptw'].dropna()))
        print('----------------------------------------------------------------------')
        print('IPW Diagnostic for positivity')
        print('''If the mean of the weights is far from either the min or max, this may\n indicate the model is 
                incorrect or positivity is violated''')
        print('Standard deviation can help in IPTW model selection')
        print('----------------------------------------------------------------------')
        print('Mean weight:\t\t\t', round(avg, decimal))
        print('Standard Deviation:\t\t', round(sd, decimal))
        print('Minimum weight:\t\t\t', round(mn, decimal))
        print('Maximum weight:\t\t\t', round(mx, decimal))
        print('----------------------------------------------------------------------')

    def StandardizedDifference(self, variable, var_type, decimal=3):
        """Calculates the standardized mean difference between the treat/exposed and untreated/unexposed for a
        specified variable. Useful for checking whether a confounder was balanced between the two treatment groups
        by the specified IPTW model

        SMD based on: Austin PC 2011; https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3144483/

        variable:
            -label for variable to calculate the standardized difference
        var_type:
            -variable type. Options are 'binary' or 'continuous'
        decimal:
            -decimal places to display in results. Default is 3
        """
        if var_type == 'binary':
            wt1 = np.sum(self.df.loc[((self.df[variable] == 1) & (self.df[self.ex] == 1))]['iptw'].dropna())
            wt2 = np.sum(self.df.loc[(self.df[self.ex] == 1)].dropna()['iptw'])
            wt = wt1 / wt2
            wn1 = np.sum(self.df.loc[(self.df[variable] == 1) & (self.df[self.ex] == 0)]['iptw'].dropna())
            wn2 = np.sum(self.df.loc[self.df[self.ex] == 0]['iptw'].dropna())
            wn = wn1 / wn2
            wsmd = (wt - wn) / math.sqrt((wt*(1 - wt) + wn*(1 - wn))/2)
        elif var_type == 'continuous':
            wmn, wmt = self._weighted_avg(self.df, v=variable, w='iptw', t=self.ex)
            wsn, wst = self._weighted_std(self.df, v=variable, w='iptw', t=self.ex, xbar=[wmn, wmt])
            wsmd = (wmt - wmn) / math.sqrt((wst**2 + wsn**2)/2)
        else:
            raise ValueError('The only variable types currently supported are binary and continuous')
        print('----------------------------------------------------------------------')
        print('IPW Diagnostic for balance: Standardized Differences')
        if var_type == 'binary':
            print('\tBinary variable: ' + variable)
        if var_type == 'continuous':
            print('\tContinuous variable: ' + variable)
        print('----------------------------------------------------------------------')
        print('Weighted SMD: \t', round(wsmd, decimal))
        print('----------------------------------------------------------------------')

    def _weight_calculator(self, df, denominator, numerator):
        """
        Calculates the IPTW based on the predicted probabilities and the specified group to standardize to
        """
        if self.stabilized:  # Stabilized weights
            if self.standardize == 'population':
                df['w'] = np.where(df[self.ex] == 1, (df[numerator] / df[denominator]),
                                   ((1 - df[numerator]) / (1 - df[denominator])))
                df.loc[(df[self.ex] != 1) & (df[self.ex] != 0), 'w'] = np.nan
            # Stabilizing to exposed (compares all exposed if they were exposed versus unexposed)
            elif self.standardize == 'exposed':
                df['w'] = np.where(df[self.ex] == 1, 1,
                                   ((df[denominator] / (1 - df[denominator])) * ((1 - df[numerator]) /
                                                                                 df[numerator])))
                df.loc[(df[self.ex] != 1) & (df[self.ex] != 0), 'w'] = np.nan
            # Stabilizing to unexposed (compares all unexposed if they were exposed versus unexposed)
            else:
                df['w'] = np.where(df[self.ex] == 1,
                                   (((1 - df[denominator]) / df[denominator]) * (df[numerator] /
                                                                                 (1 - df[numerator]))),
                                   1)
                df.loc[(df[self.ex] != 1) & (df[self.ex] != 0), 'w'] = np.nan

        else:  # Unstabilized weights
            if self.standardize == 'population':
                df['w'] = np.where(df[self.ex] == 1, 1 / df[denominator], 1 / (1 - df[denominator]))
                df.loc[(df[self.ex] != 1) & (df[self.ex] != 0), 'w'] = np.nan
            # Stabilizing to exposed (compares all exposed if they were exposed versus unexposed)
            elif self.standardize == 'exposed':
                df['w'] = np.where(df[self.ex] == 1, 1, (df[denominator] / (1 - df[denominator])))
                df.loc[(df[self.ex] != 1) & (df[self.ex] != 0), 'w'] = np.nan
            # Stabilizing to unexposed (compares all unexposed if they were exposed versus unexposed)
            else:
                df['w'] = np.where(df[self.ex] == 1, ((1 - df[denominator]) / df[denominator]), 1)
                df.loc[(df[self.ex] != 1) & (df[self.ex] != 0), 'w'] = np.nan
        return df['w']

    @staticmethod
    def _weighted_avg(df, v, w, t):
        """
        Calculates the weighted mean for continuous variables
        """
        l = []
        for i in [0, 1]:
            n = sum(df.loc[df[t] == i][v] * df.loc[df[t] == i][w])
            d = sum(df.loc[df[t] == i][w])
            a = n / d
            l.append(a)
        return l[0], l[1]

    @staticmethod
    def _weighted_std(df, v, w, t, xbar):
        """
        Calculates the weighted standard deviation for continuous variables
        """
        l = []
        for i in [0, 1]:
            n1 = sum(df.loc[df[t] == i][w])
            d1 = sum(df.loc[df[t] == i][w]) ** 2
            d2 = sum(df.loc[df[t] == i][w] ** 2)
            n2 = sum(df.loc[df[t] == i][w] * ((df.loc[df[t] == i][v] - xbar[i]) ** 2))
            a = ((n1 / (d1 - d2)) * n2)
            l.append(a)
        return l[0], l[1]



import warnings
import math 
import numpy as np

def trapezoidal(mini, mode1, mode2, maxi, size=None):
    '''Creates trapezoidal distribution based on Fox & Lash 2005. This function 
    can be used to generate distributions of probabilities and effect measures for
    sensitivity analyses. It is particularly useful when used in conjunction with 
    rr_corr to determine a distribution of potential results due to a single unadjusted
    confounder
    
    mini:
        -minimum value of trapezoidal distribution
    mode1:
        -Start of uniform distribution
    mode2:
        -End of uniform distribution
    maxi:
        -maximum value of trapezoidal distribution
    size:
        -number of observations to generate. Default is None, which returns a single draw
    
    Example)
    >>>zepid.sens_analysis.trapezoidal(mini=0.2,mode1=0.3,mode2=0.5,maxi=0.6)
    '''
    if size == None:
        p = np.random.uniform()
        v = (p*(maxi+mode2-mini-mode1)+(mini+mode1)) / 2
        if v < mode1:
            v = mini + np.sqrt((mode1-mini)*(2*v-mini-mode1))
        elif v > mode2:
            v = maxi - np.sqrt(2*(maxi-mode2)*(v-mode2))
        else:
            pass   
        return v
    
    #draws the number specified by size
    elif type(size) == int:
        va = []
        for i in range(size):
            va.append(trapezoidal(mini=mini,mode1=mode1,mode2=mode2,maxi=maxi,size=None))
        return va
    
    #ValueError is size is not an integer        
    else:
        raise ValueError('"size" must be an integer')
        



import warnings
import math 
import numpy as np
import pandas as pd
from scipy import stats
import matplotlib.pyplot as plt

class MonteCarloRR:
    '''Monte Carlo simulation to assess the impact of an unmeasured binary confounder on the results
    of a study. Observed RR comes from the data analysis, while the RR between the unmeasured confounder
    and the outcome should be obtained from prior literature or constitute an reasonable guess.
    Probability of exposure between the groups should also be reasonable numbers. 
    
    observed_RR:
        -observed RR from the data, not accounting for some binary unmeasured confounder
    sample:
        -number of MC simulations to run. It is important that the specified size of later distributions
         matches this number of samples
    
    Example)
    >>>from zepid.sensitivity_analysis import MonteCarloRR
    >>>mcrr = MonteCarloRR(observed_RR=)
    >>>
    >>>
    >>>
    >>>mcrr.fit(sample=10000)
    '''
    def __init__(self,observed_RR,sample):
        self.RRo = observed_RR
        self._rrdist = False
        self._p1dist = False
        self._p0dist = False
        self.sample = sample

    def confounder_RR_distribution(self,dist,seed=None):
        '''Distribution of the risk ratio between the unmeasured confounder and the outcome. This
        value should come from prior literature or a reasonable guess. Any numpy random distribution
        can be based to this function. Alternatively, the trapezoid distribution within this library
        can also be used

        dist:
            -distribution from which the confounder-outcome Risk Ratio is pulled from. Input should be
             something like numpy.random.triangular(left=0.9,mode=1.2,right=1.6)
        '''
        if seed != None:
            np.random.seed(seed)
        self.RRc_dist = dist
        self._rrdist = True

    def prop_confounder_exposed(self,dist,seed=None):
        '''Distribution of the proportion of the unmeasured confounder in the exposed group. This
        value should come from prior literature or a reasonable guess. Any numpy random distribution
        can be based to this function. Alternatively, the trapezoid distribution within this library
        can also be used

        dist:
            -distribution from which the confounder-outcome Risk Ratio is pulled from. Input should be
             something like numpy.random.triangular(left=0.1,mode=0.2,right=0.3)
        '''
        if seed != None:
            np.random.seed(seed)
        self.pc1 = dist
        self._p1dist = True

    def prop_confounder_unexposed(self,dist,seed=None):
        '''Distribution of the proportion of the unmeasured confounder in the unexposed group. This
        value should come from prior literature or a reasonable guess. Any numpy random distribution
        can be based to this function. Alternatively, the trapezoid distribution within this library
        can also be used

        dist:
            -distribution from which the confounder-outcome Risk Ratio is pulled from. Input should be
             something like numpy.random.triangular(left=0.2,mode=0.3,right=0.4)
        '''
        if seed != None:
            np.random.seed(seed)
        self.pc0 = dist
        self._p0dist = True

    def fit(self):
        '''After the observed Risk Ratio, distribution of the confounder-outcome Risk Ratio, proportion
        of the unmeasured confounder in exposed, proportion of the unmeasured confounder in the unexposed.

        sample:
            -number of samples to pull from each distribution and recalculate the corrected risk ratio. The
             default is 10,000. A high number is generally recommended
        seed:
            -set the numpy seed. Default is None

        Formula:
                    RR* = RR / d
                      d = (p1 * (RRc-1) + 1) / (p0 * (RRc - 1) + 1)
        RR* : corrected risk ratio
        RR  : observed risk ratio in the data set
        RRc : risk ratio between unmeasured confounder and outcome
        p1  : probability/proportion of unmeasured confounder in exposed
        p0  : probability/proportion of unmeasured confounder in unexposed
        '''
        if self._rrdist == False:
            raise ValueError('"confounder_RR_distribution()" has not been specified')
        if self._p1dist == False:
            raise ValueError('"prop_confounder_exposed()" has not been specified')
        if self._p0dist == False:
            raise ValueError('"prop_confounder_unexposed()" has not been specified')
        
        #Monte Carlo
        sf = pd.DataFrame(index = [i for i in range(self.sample)])
        sf['RR_cnf'] = self.RRc_dist
        sf['p1'] = self.pc1
        sf['p0'] = self.pc0
        sf['RR_obs'] = self.RRo
        sf['RR_cor'] = sf['RR_obs'] / ((sf['p1']*(sf['RR_cnf']-1)+1) / ((sf['p0']*(sf['RR_cnf']-1)+1)))

        #setting new attribute
        self.corrected_RR = list(sf['RR_cor'])

    def summary(self,decimal=4):
        '''Generate the summary information after the corrected risk ratio distribution is
        generated. fit() must be run before this

        decimal:
            -number of decimal places to display in output. Default is 4
        '''
        print('----------------------------------------------------------------------')
        print('Median corrected Risk Ratio: ',round(np.median(self.corrected_RR),decimal))
        print('Mean corrected Risk Ratio: ',round(np.mean(self.corrected_RR),decimal))
        print('25th & 75th Percentiles: ',np.round_(np.percentile(self.corrected_RR,q=[25,75]),decimals=decimal))     
        print('2.5th & 97.5th Percentiles: ',np.round_(np.percentile(self.corrected_RR,q=[2.5,97.5]),decimals=decimal))     
        print('----------------------------------------------------------------------')

    def plot(self,bw_method='scott',fill=True,color='b'):
        '''Generate a Gaussian kernel density plot of the corrected risk ratio distribution. The
        kernel density used is SciPy's Gaussian kernel. Either Scott's Rule or Silverman's Rule can
        be implemented

        bw_method:
            -method used to estimate the bandwidth. Following SciPy, either 'scott' or 'silverman' are valid options
        fill:
            -whether to color the area under the density curves. Default is true
        color:
            -color of the line/area for the treated group. Default is Blue
        '''
        x = np.linspace(np.min(self.corrected_RR),np.max(self.corrected_RR),100)
        density = stats.kde.gaussian_kde(self.corrected_RR,bw_method=bw_method)

        #Creating density plot
        ax = plt.gca()
        if fill == True:
            ax.fill_between(x,density(x),color=color,alpha=0.2)
        ax.plot(x, density(x),color=color)
        ax.set_xlabel('Corrected Risk Ratio')
        ax.set_ylabel('Density')
        return ax



# utilities for IPW methods. Basically for propensity_score(), which they all use

import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.genmod.families import links

def propensity_score(df, model, mresult=True):
    '''Generate propensity scores (probability) based on the model input. Uses logistic regression model
    to calculate

    Returns fitted propensity score model

    df:
        -Dataframe for the model
    model:
        -Model to fit the logistic regression to. Example) 'y ~ var1 + var2'
    mresult:
        -Whether to print the logistic regression results. Default is True

    Example)
    >>>zepid.ipw.propensity_score(df=data,model='X ~ Z + var1 + var2')
    '''
    f = sm.families.family.Binomial(sm.families.links.logit)
    log = smf.glm(model, df, family=f).fit()
    if mresult == True:
        print('\n----------------------------------------------------------------')
        print('MODEL: ' + model)
        print('-----------------------------------------------------------------')
        print(log.summary())
    return log



import warnings
import math
from tabulate import tabulate
from scipy.stats import norm


def risk_ci(events, total, alpha=0.05, decimal=3, confint='wald'):
    """Calculate two-sided (1-alpha)% Confidence interval of Risk. Note
    relies on the Central Limit Theorem, so there must be at least 5 events
    and 5 nonevents

    events:
        -Number of events/outcomes that occurred
    total:
        -total number of subjects that could have experienced the event
    alpha:
        -Alpha level. Default is 0.05
    decimal:
        -Number of decimal places to display. Default is 3 decimal places
    confint:
        -Type of confidence interval to generate. Current options are
            wald
            hypergeometric
    """
    risk = events / total
    c = 1 - alpha / 2
    zalpha = norm.ppf(c, loc=0, scale=1)
    if confint == 'wald':
        lr = math.log(risk / (1 - risk))
        sd = math.sqrt((1 / events) + (1 / (total - events)))
        lower = 1 / (1 + math.exp(-1 * (lr - zalpha * sd)))
        upper = 1 / (1 + math.exp(-1 * (lr + zalpha * sd)))
    elif confint == 'hypergeometric':
        sd = math.sqrt(events * (total - events) / (total ** 2 * (total - 1)))
        lower = risk - zalpha * sd
        upper = risk + zalpha * sd
    else:
        raise ValueError('Please specify a valid confidence interval')
    print('Risk: ' + str(round(risk, decimal)) + ', ', str(round(100 * (1 - alpha), 1)) + '% CI: (' +
          str(round(lower, decimal)) + ', ' + str(round(upper, decimal)) + ')')


def ir_ci(events, time, alpha=0.05, decimal=3):
    """Calculate two-sided (1-alpha)% Wald Confidence interval of Incidence Rate

    events:
        -number of events/outcomes that occurred
    time:
        -total person-time contributed in this group
    alpha:
        -alpha level. Default is 0.05
    decimal:
        -amount of decimal places to display. Default is 3 decimal places
    """
    c = 1 - alpha / 2
    ir = events / time
    zalpha = norm.ppf(c, loc=0, scale=1)
    se = math.sqrt(events / (time ** 2))
    lower = ir - zalpha * se
    upper = ir + zalpha * se
    print('Incidence rate: ' + str(round(ir, decimal)) + ', ' + str(round(100 * (1 - alpha), 1)) + '% CI: (' +
          str(round(lower, decimal)) + ', ' + str(round(upper, decimal)) + ')')


def rr(a, b, c, d, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """Calculates the Risk Ratio from count data.

    a:
        -count of exposed individuals with outcome
    b:
        -count of unexposed individuals with outcome
    c:
        -count of exposed individuals without outcome
    d:
        -count of unexposed individuals without outcome
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True, which prints the results
    return_result:
        -Whether to return the RR as a object. Default is False
    """
    if (a < 0) or (b < 0) or (c < 0) or (d < 0):
        raise ValueError('All numbers must be positive')
    if (a <= 5) or (b <= 5) or (c <= 5) or (d <= 5):
        warnings.warn('At least one cell count is less than 5, therefore confidence interval approximation is invalid')
    zalpha = norm.ppf((1 - alpha / 2), loc=0, scale=1)
    r1 = a / (a + b)
    r0 = c / (c + d)
    relrisk = r1 / r0
    SE = math.sqrt((1 / a) - (1 / (a + b)) + (1 / c) - (1 / (c + d)))
    lnrr = math.log(relrisk)
    lcl = lnrr - (zalpha * SE)
    ucl = lnrr + (zalpha * SE)
    if print_result:
        print(tabulate([['E=1', a, b], ['E=0', c, d]], headers=['', 'D=1', 'D=0'], tablefmt='grid'))
        print('----------------------------------------------------------------------')
        print('Exposed')
        risk_ci(a, a + b, alpha=alpha, decimal=decimal)
        print('Unexposed')
        risk_ci(c, c + d, alpha=alpha, decimal=decimal)
        print('----------------------------------------------------------------------')
        print('Risk Ratio:', round(relrisk, decimal))
        print(str(round(100 * (1 - alpha), 1)) + '% two-sided CI: (' + str(round(math.exp(lcl), decimal)), ',',
              str(round(math.exp(ucl), decimal)) + ')')
        print('Confidence Limit Ratio: ', round(((math.exp(ucl)) / (math.exp(lcl))), decimal))
        print('Standard Deviation: ', round(SE, decimal))
        print('----------------------------------------------------------------------\n')
    if return_result:
        return relrisk


def rd(a, b, c, d, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """Calculates the Risk Difference from count data.

    a:
        -count of exposed individuals with outcome
    b:
        -count of unexposed individuals with outcome
    c:
        -count of exposed individuals without outcome
    d:
        -count of unexposed individuals without outcome
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the RR as a object. Default is False
    """
    if (a < 0) or (b < 0) or (c < 0) or (d < 0):
        raise ValueError('All numbers must be positive')
    if (a <= 5) or (b <= 5) or (c <= 5) or (d <= 5):
        warnings.warn('At least one cell count is less than 5, therefore confidence interval approximation is invalid')
    zalpha = norm.ppf((1 - alpha / 2), loc=0, scale=1)
    r1 = a / (a + b)
    r0 = c / (c + d)
    riskdiff = r1 - r0
    SE = math.sqrt(((a * b) / ((a + b) ** 2 * (a + b - 1))) + ((c * d) / (((c + d) ** 2) * (c + d - 1))))
    lcl = riskdiff - (zalpha * SE)
    ucl = riskdiff + (zalpha * SE)
    if print_result:
        print(tabulate([['E=1', a, b], ['E=0', c, d]], headers=['', 'D=1', 'D=0'], tablefmt='grid'))
        print('----------------------------------------------------------------------')
        print('Exposed')
        risk_ci(a, a + b, alpha=alpha, decimal=decimal)
        print('Unexposed')
        risk_ci(c, c + d, alpha=alpha, decimal=decimal)
        print('----------------------------------------------------------------------')
        print('Risk Difference:', round(riskdiff, decimal))
        print(str(round(100 * (1 - alpha), 1)) + '%  two-sided CI: (', round(lcl, decimal), ', ', round(ucl, decimal),
              ')')
        print('Confidence Limit Difference: ', round((ucl - lcl), decimal))
        print('Standard Deviation: ', round(SE, decimal))
        print('----------------------------------------------------------------------\n')
    if return_result:
        return riskdiff


def nnt(a, b, c, d, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """Calculates the Number Needed to Treat from count data

    a:
        -count of exposed individuals with outcome
    b:
        -count of unexposed individuals with outcome
    c:
        -count of exposed individuals without outcome
    d:
        -count of unexposed individuals without outcome
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the RR as a object. Default is False
    """
    if (a < 0) or (b < 0) or (c < 0) or (d < 0):
        raise ValueError('All numbers must be positive')
    if (a <= 5) or (b <= 5) or (c <= 5) or (d <= 5):
        warnings.warn('At least one cell count is less than 5, therefore confidence interval approximation is invalid')
    zalpha = norm.ppf((1 - alpha / 2), loc=0, scale=1)
    ratiod1 = a / (a + b)
    ratiod2 = c / (c + d)
    riskdiff = ratiod1 - ratiod2
    SE = math.sqrt(((a * b) / ((a + b)**2 * (a + b - 1))) + ((c * d) / (((c + d) ** 2) * (c + d - 1))))
    lcl_rd = (riskdiff - (zalpha * SE))
    ucl_rd = (riskdiff + (zalpha * SE))
    try:
        NNT = 1 / riskdiff
    except:
        NNT = 'inf'
    try:
        ucl = 1 / lcl_rd
    except:
        ucl = 'inf'
    try:
        lcl = 1 / ucl_rd
    except:
        lcl = 'inf'
    if print_result:
        print('----------------------------------------------------------------------')
        print('Risk Difference: ', round(riskdiff, decimal))
        print('----------------------------------------------------------------------')
        if riskdiff == 0:
            print('Number Needed to Treat = infinite')
        else:
            if riskdiff > 0:
                print('Number Needed to Harm: ', round(abs(NNT), decimal), '\n')
            if riskdiff < 0:
                print('Number Needed to Treat: ', round(abs(NNT), decimal), '\n')
        print(str(round(100 * (1 - alpha), 1)) + '% two-sided CI: ')
        if lcl_rd < 0 < ucl_rd:
            print('NNH ', round(abs(lcl), decimal), 'to infinity to NNT ', round(abs(ucl), decimal))
        elif 0 < lcl_rd:
            print('NNT ', round(abs(lcl), decimal), ' to ', round(abs(ucl), decimal))
        else:
            print('NNH ', round(abs(lcl), decimal), ' to ', round(abs(ucl), decimal))
    print('----------------------------------------------------------------------\n')
    if return_result:
        return nnt


def oddsratio(a, b, c, d, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """Calculates the Odds Ratio from count data

    a:
        -count of exposed individuals with outcome
    b:
        -count of unexposed individuals with outcome
    c:
        -count of exposed individuals without outcome
    d:
        -count of unexposed individuals without outcome
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the RR as a object. Default is False
    """
    if (a < 0) or (b < 0) or (c < 0) or (d < 0):
        raise ValueError('All numbers must be positive')
    if (a <= 5) or (b <= 5) or (c <= 5) or (d <= 5):
        warnings.warn('At least one cell count is less than 5, therefore confidence interval approximation is invalid')
    zalpha = norm.ppf((1 - alpha / 2), loc=0, scale=1)
    or1 = a / b
    or0 = c / d
    oddsr = or1 / or0
    SE = math.sqrt((1 / a) + (1 / b) + (1 / c) + (1 / d))
    lnor = math.log(oddsr)
    lcl = lnor - (zalpha * SE)
    ucl = lnor + (zalpha * SE)
    if print_result:
        print(tabulate([['E=1', a, b], ['E=0', c, d]], headers=['', 'D=1', 'D=0'], tablefmt='grid'))
        print('----------------------------------------------------------------------')
        print('Odds exposed:', round(or1, decimal))
        print('Odds unexposed:', round(or0, decimal))
        print('----------------------------------------------------------------------')
        print('Odds Ratio:', round(oddsr, decimal))
        print(str(round(100 * (1 - alpha), 1)) + '% two-sided CI: (', round(math.exp(lcl), decimal), ', ',
              round(math.exp(ucl), decimal), ')')
        print('Confidence Limit Ratio: ', round(((math.exp(ucl)) / (math.exp(lcl))), decimal))
        print('Standard Deviation: ', round(SE, decimal))
        print('----------------------------------------------------------------------\n')
    if return_result:
        return oddsr


def irr(a, c, t1, t2, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """Calculates the Incidence Rate Ratio from count data

    a:
        -count of exposed with outcome
    b:
        -count of unexposed with outcome
    T1:
        -person-time contributed by those who were exposed
    T2:
        -person-time contributed by those who were unexposed
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the RR as a object. Default is False
    """
    if (a < 0) or (c < 0) or (t1 <= 0) | (t2 <= 0):
        raise ValueError('All numbers must be positive')
    if (a <= 5) or (c <= 5):
        warnings.warn('At least one event count is less than 5, therefore confidence interval approximation is invalid')
    zalpha = norm.ppf((1 - alpha / 2), loc=0, scale=1)
    irate1 = a / t1
    irate2 = c / t2
    irr = irate1 / irate2
    SE = math.sqrt((1 / a) + (1 / c))
    lnirr = math.log(irr)
    lcl = lnirr - (zalpha * SE)
    ucl = lnirr + (zalpha * SE)
    if print_result:
        print(tabulate([['E=1', a, t1], ['E=0', c, t2]], headers=['', 'D=1', 'Person-time'], tablefmt='grid'))
        print('----------------------------------------------------------------------')
        print('Exposed')
        ir_ci(a, t1, alpha=alpha, decimal=decimal)
        print('Unexposed')
        ir_ci(c, t2, alpha=alpha, decimal=decimal)
        print('----------------------------------------------------------------------')
        print('Incidence Rate Ratio:', round(irr, decimal))
        print(str(round(100 * (1 - alpha), 1)) + '% two-sided CI: (', round(math.exp(lcl), decimal), ', ',
              round(math.exp(ucl), decimal), ')')
        print('Confidence Limit Ratio: ', round(((math.exp(ucl)) / (math.exp(lcl))), decimal))
        print('Standard Deviation: ', round(SE, decimal))
        print('----------------------------------------------------------------------\n')
    if return_result:
        return irr


def ird(a, c, t1, t2, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """Calculates the Incidence Rate Difference from count data

    a:
        -count of exposed with outcome
    b:
        -count of unexposed with outcome
    T1:
        -person-time contributed by those who were exposed
    T2:
        -person-time contributed by those who were unexposed
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the RR as a object. Default is False
    """
    if (a < 0) or (c < 0) or (t1 <= 0) or (t2 <= 0):
        raise ValueError('All numbers must be positive')
    if (a <= 5) or (c <= 5):
        warnings.warn('At least one event count is less than 5, therefore confidence interval approximation is invalid')
    zalpha = norm.ppf((1 - alpha / 2), loc=0, scale=1)
    rated1 = a / t1
    rated2 = c / t2
    ird = rated1 - rated2
    SE = math.sqrt((a / (t1**2)) + (c / (t2**2)))
    lcl = ird - (zalpha * SE)
    ucl = ird + (zalpha * SE)
    if print_result:
        print(tabulate([['E=1', a, t1], ['E=0', c, t2]], headers=['', 'D=1', 'Person-time'], tablefmt='grid'))
        print('----------------------------------------------------------------------')
        print('Exposed')
        ir_ci(a, t1, alpha=alpha, decimal=decimal)
        print('Unexposed')
        ir_ci(c, t2, alpha=alpha, decimal=decimal)
        print('----------------------------------------------------------------------')
        print('Incidence Rate Difference:', round(ird, decimal))
        print(str(round(100 * (1 - alpha), 1)) + '% two-sided CI: (', round(lcl, decimal), ', ', round(ucl, decimal),
              ')')
        print('Confidence Limit Difference: ', round((ucl - lcl), decimal))
        print('Standard Deviation: ', round(SE, decimal))
        print('----------------------------------------------------------------------\n')
    if return_result:
        return ird


def acr(a, b, c, d, decimal=3):
    """Calculates the estimated Attributable Community Risk (ACR) from count data. ACR is also
    known as Population Attributable Risk. Since this is commonly confused with the population
    attributable fraction, the name ACR is used to clarify differences in the formulas

    a:
        -count of exposed individuals with outcome
    b:
        -count of unexposed individuals with outcome
    c:
        -count of exposed individuals without outcome
    d:
        -count of unexposed individuals without outcome
    decimal:
        -amount of decimal points to display. Default is 3
    """
    if (a < 0) or (b < 0) or (c < 0) or (d < 0):
        raise ValueError('All numbers must be positive')
    rt = (a + c) / (a + b + c + d)
    r0 = c / (c + d)
    acr = (rt - r0)
    print(tabulate([['E=1', a, b], ['E=0', c, d]], headers=['', 'D=1', 'D=0'], tablefmt='grid'))
    print('----------------------------------------------------------------------')
    print('ACR: ', round(acr, decimal))
    print('----------------------------------------------------------------------\n')


def paf(a, b, c, d, decimal=3):
    """Calculates the Population Attributable Fraction from count data

    a:
        -count of exposed individuals with outcome
    b:
        -count of unexposed individuals with outcome
    c:
        -count of exposed individuals without outcome
    d:
        -count of unexposed individuals without outcome
    decimal:
        -amount of decimal points to display. Default is 3
    """
    if (a < 0) or (b < 0) or (c < 0) or (d < 0):
        raise ValueError('All numbers must be positive')
    rt = (a + c) / (a + b + c + d)
    r0 = c / (c + d)
    paf = (rt - r0) / rt
    print(tabulate([['E=1', a, b], ['E=0', c, d]], headers=['', 'D=1', 'D=0'], tablefmt='grid'))
    print('----------------------------------------------------------------------')
    print('PAF: ', round(paf, decimal))
    print('----------------------------------------------------------------------\n')


def prop_to_odds(prop):
    """Convert proportion to odds. Returns the corresponding odds

    prop:
        -proportion that is desired to transform into odds
    """
    odds = prop / (1 - prop)
    return odds


def odds_to_prop(odds):
    """Convert odds to proportion. Returns the corresponding proportion

    odds:
        -odds that is desired to transform into a proportion
    """
    prop = odds / (1 + odds)
    return prop


def counternull_pvalue(estimate, lcl, ucl, sided='two', alpha=0.05, decimal=3):
    """Calculates the counternull based on Rosenthal R & Rubin DB (1994). It is useful to prevent over-interpretation
    of results. For a full discussion and how to interpret the estimate and p-value, see Rosenthal & Rubin.

    Warning: Make sure that the confidence interval points put into
    the equation match the alpha level calculation

    estimate:
        -Point estimate for result
    lcl:
        -Lower confidence limit
    ucl:
        -Upper confidence limit
    sided:
        -Whether to compute the upper one-sided, lower one-sided, or two-sided counternull
         p-value. Default is the two-sided
            'upper'     Upper one-sided p-value
            'lower'     Lower one-sided p-value
            'two'       Two-sided p-value
    alpha:
        -Alpha level for p-value. Default is 0.05. Verify that this is the same alpha used to
         generate confidence intervals
    decimal:
        -Number of decimal places to display. Default is three
    """
    zalpha = norm.ppf((1 - alpha / 2), loc=0, scale=1)
    se = (ucl - lcl) / (zalpha * 2)
    cnull = 2 * estimate
    up_cn = norm.cdf(x=cnull, loc=estimate, scale=se)
    lp_cn = 1 - up_cn
    lowerp = norm.cdf(x=estimate, loc=cnull, scale=se)
    upperp = 1 - lowerp
    twosip = 2 * min([up_cn, lp_cn])
    print('----------------------------------------------------------------------')
    print('Alpha = ', alpha)
    print('----------------------------------------------------------------------')
    print('Counternull estimate = ', cnull)
    if sided == 'upper':
        print('Upper one-sided counternull p-value: ', round(upperp, decimal))
    elif sided == 'lower':
        print('Lower one-sided counternull p-value: ', round(lowerp, decimal))
    else:
        print('Two-sided counternull p-value: ', round(twosip, decimal))
    print('----------------------------------------------------------------------\n')


def semibayes(prior_mean, prior_lcl, prior_ucl, mean, lcl, ucl, ln_transform=False, alpha=0.05, decimal=3):
    """A simple Bayesian Analysis. Note that this analysis assumes normal distribution for the
    continuous measure. See chapter 18 of Modern Epidemiology 3rd Edition (specifically pages 334, 340)

    Warning: Make sure that the alpha used to generate the confidence intervals matches the alpha
    used in this calculation. Additionally, this calculation can only handle normally distributed
    priors and observed

    prior_mean:
        -Prior designated point estimate
    prior_lcl:
        -Prior designated lower confidence limit
    prior_ucl:
        -Prior designated upper confidence limit
    mean:
        -Point estimate result obtained from analysis
    lcl:
        -Lower confidence limit estimate obtained from analysis
    ucl:
        -Upper confidence limit estimate obtained from analysis
    ln_transform:
        -Whether to natural log transform results before conducting analysis. Should be used for
         RR, OR, or or other Ratio measure. Default is False (use for RD and other absolute measures)
    alpha:
        -Alpha level for confidence intervals. Default is 0.05
    decimal:
        -Number of decimal places to display. Default is three
    """
    # Transforming to log scale if ratio measure
    if ln_transform:
        prior_mean = math.log(prior_mean)
        prior_lcl = math.log(prior_lcl)
        prior_ucl = math.log(prior_ucl)
        mean = math.log(mean)
        lcl = math.log(lcl)
        ucl = math.log(ucl)
    zalpha = norm.ppf((1 - alpha / 2), loc=0, scale=1)

    # Extracting prior SD
    prior_sd = (prior_ucl - prior_lcl) / (2 * zalpha)
    prior_var = prior_sd ** 2
    prior_w = 1 / prior_var

    # Extracting observed SD
    sd = (ucl - lcl) / (2 * zalpha)
    var = sd ** 2
    w = 1 / var

    # Checking Prior
    check = (mean - prior_mean) / ((var + prior_var) ** (1 / 2))
    # TODO add some logic to this part to generate a warning when necessary

    # Calculating posterior
    post_mean = ((prior_mean * prior_w) + (mean * w)) / (prior_w + w)
    post_var = 1 / (prior_w + w)
    sd = math.sqrt(post_var)
    post_lcl = post_mean - zalpha * sd
    post_ucl = post_mean + zalpha * sd

    # Transforming back if ratio measure
    if ln_transform:
        post_mean = math.exp(post_mean)
        post_lcl = math.exp(post_lcl)
        post_ucl = math.exp(post_ucl)
        prior_mean = math.exp(prior_mean)
        prior_lcl = math.exp(prior_lcl)
        prior_ucl = math.exp(prior_ucl)
        mean = math.exp(mean)
        lcl = math.exp(lcl)
        ucl = math.exp(ucl)

    # Presenting Results
    print('----------------------------------------------------------------------')
    print('Prior Estimate: ', round(prior_mean, decimal))
    print(str(round((1 - alpha) * 100, 1)) + '% Prior Confidence Interval: (', round(prior_lcl, decimal), ', ',
          round(prior_ucl, decimal), ')')
    print('----------------------------------------------------------------------')
    print('Point Estimate: ', round(mean, decimal))
    print(str(round((1 - alpha) * 100, 1)) + '% Confidence Interval: (', round(lcl, decimal), ', ', round(ucl, decimal),
          ')')
    print('----------------------------------------------------------------------')
    print('Posterior Estimate: ', round(post_mean, decimal))
    print(str(round((1 - alpha) * 100, 1)) + '% Posterior Probability Interval: (', round(post_lcl, decimal), ', ',
          round(post_ucl, decimal), ')')
    print('----------------------------------------------------------------------\n')


def sensitivity(detected, cases, alpha=0.05, decimal=3, confint='wald', print_result=True, return_result=False):
    """
    Calculate the Sensitivity from number of detected cases and the number of total true cases.

    detected:
        -number of true cases detected via testing criteria
    cases:
        -total number of true/actual cases
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the RR as a object. Default is False
    """
    if (detected < 0) or (cases < 0):
        raise ValueError('All numbers must be positive')
    if detected > cases:
        raise ValueError('Detected true cases must be less than or equal to the total number of cases')
    if cases <= 5:
        warnings.warn('At least one cell count is less than 5, therefore confidence interval approximation is invalid')
    sens = detected / cases
    zalpha = norm.ppf(1 - alpha / 2, loc=0, scale=1)
    if confint == 'wald':
        ls = math.log(sens / (1 - sens))
        sd = math.sqrt((1 / detected) + (1 / (cases - detected)))
        lower = 1 / (1 + math.exp(-1 * (ls - zalpha * sd)))
        upper = 1 / (1 + math.exp(-1 * (ls + zalpha * sd)))
    elif confint == 'hypergeometric':
        sd = math.sqrt(detected * (cases - detected) / (cases ** 2 * (cases - 1)))
        lower = sens - zalpha * sd
        upper = sens + zalpha * sd
    else:
        raise ValueError('Please specify a valid confidence interval')
    if print_result:
        print('Sensitivity: ' + str(round(sens, decimal)) + ', ', str(round(100 * (1 - alpha), 1)) + '% CI: (' +
              str(round(lower, decimal)) + ', ' + str(round(upper, decimal)) + ')')
    if return_result:
        return sens


def specificity(detected, noncases, alpha=0.05, decimal=3, confint='wald', print_result=True, return_result=False):
    """
    Calculate the Sensitivity from number of detected cases and the number of total true cases.

    detected:
        -number of false cases detected via testing criteria
    cases:
        -total number of true/actual noncases
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    confint:
        -type of confidence interval to generate
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the RR as a object. Default is False
    """
    if (detected < 0) or (noncases < 0):
        raise ValueError('All numbers must be positive')
    if detected > noncases:
        raise ValueError('Detected true cases must be less than or equal to the total number of cases')
    if noncases <= 5:
        warnings.warn('At least one cell count is less than 5, therefore confidence interval approximation is invalid')
    spec = 1 - (detected / noncases)
    zalpha = norm.ppf(1 - alpha / 2, loc=0, scale=1)
    if confint == 'wald':
        ls = math.log(spec / (1 - spec))
        sd = math.sqrt((1 / detected) + (1 / (noncases - detected)))
        lower = 1 / (1 + math.exp(-1 * (ls - zalpha * sd)))
        upper = 1 / (1 + math.exp(-1 * (ls + zalpha * sd)))
    elif confint == 'hypergeometric':
        sd = math.sqrt(detected * (noncases - detected) / (noncases ** 2 * (cases - 1)))
        lower = spec - zalpha * sd
        upper = spec + zalpha * sd
    else:
        raise ValueError('Please specify a valid confidence interval')
    if print_result:
        print('Specificity: ' + str(round(spec, decimal)) + ', ', str(round(100 * (1 - alpha), 1)) + '% CI: (' +
              str(round(lower, decimal)) + ', ' + str(round(upper, decimal)) + ')')
    if return_result:
        return spec


def diagnostics(a, b, c, d, alpha=0.05, decimal=3, confint='wald', print_result=True, return_result=False):
    """
    Calculate the diagnostic criteria (sensitivity and specificity) for summary data

    a:
        -count of true cases with a positive test
    b:
        -count of true cases with a negative test
    c:
        -count of true non-cases with a positive test
    d:
        -count of true non-cases with a negative test
    alpha:
        -alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -whether to print the results. Default is True
    return_result:
        -whether to return the calculated sensitivity and specificity as a tuple. Default is False
    """
    if (a < 0) or (b < 0) or (c < 0) or (d < 0):
        raise ValueError('All numbers must be positive')
    if print_result:
        print(tabulate([['T+', a, b], ['T-', c, d]], headers=['', 'D+', 'D-'], tablefmt='grid'))
        print('----------------------------------------------------------------------')
        sensitivity(a, a+b, alpha=alpha, decimal=decimal, confint=confint, print_result=True)
        print('----------------------------------------------------------------------')
        specificity(c, c+d, alpha=alpha, decimal=decimal, confint=confint, print_result=True)
        print('----------------------------------------------------------------------')
    if return_result:
        se = sensitivity(a, a+b, print_result=False, return_result=True)
        sp = specificity(c, c+d, print_result=False, return_result=True)
        return se, sp


def ppv_converter(sensitivity, specificity, prevalence):
    """Generates the Positive Predictive Value from designated Sensitivity, Specificity, and Prevalence.
    Returns the positive predictive value

    sensitivity:
        -sensitivity of the criteria
    specificity:
        -specificity of the criteria
    prevalence:
        -prevalence of the outcome in the population
    """
    if (sensitivity > 1) or (specificity > 1) or (prevalence > 1):
        raise ValueError('sensitivity/specificity/prevalence cannot be greater than 1')
    if (sensitivity < 0) or (specificity < 0) or (prevalence < 0):
        raise ValueError('sensitivity/specificity/prevalence cannot be less than 0')
    sens_prev = sensitivity * prevalence
    nspec_nprev = (1 - specificity) * (1 - prevalence)
    ppv = sens_prev / (sens_prev + nspec_nprev)
    return ppv


def npv_converter(sensitivity, specificity, prevalence):
    """Generates the Negative Predictive Value from designated Sensitivity, Specificity, and Prevalence.
    Returns the negative predictive value

    sensitivity:
        -sensitivity of the criteria
    specificity:
        -specificity of the criteria
    prevalence:
        -prevalence of the outcome in the population
    """
    if (sensitivity > 1) or (specificity > 1) or (prevalence > 1):
        raise ValueError('sensitivity/specificity/prevalence cannot be greater than 1')
    if (sensitivity < 0) or (specificity < 0) or (prevalence < 0):
        raise ValueError('sensitivity/specificity/prevalence cannot be less than 0')
    spec_nprev = specificity * (1 - prevalence)
    nsens_prev = (1 - sensitivity) * prevalence
    npv = spec_nprev / (spec_nprev + nsens_prev)
    return npv


def screening_cost_analyzer(cost_miss_case, cost_false_pos, prevalence, sensitivity, specificity, population=10000,
                            decimal=3):
    """Compares the cost of sensivitiy/specificity of screening criteria to treating the entire population
    as test-negative and test-positive. The lowest per capita cost is considered the ideal choice. Note that
    this function only provides relative costs

    WARNING: When calculating costs, be sure to consult experts in health policy or related fields.  Costs
    should encompass more than just monetary costs, like relative costs (regret, disappointment, stigma,
    disutility, etc.). Careful consideration of relative costs between false positive and false negatives
    needs to be considered.

    cost_miss_case:
        -The relative cost of missing a case, compared to false positives. In general, set this to 1 then
         change the value under 'cost_false_pos' to reflect the relative cost
    cost_false_pos:
        -The relative cost of a false positive case, compared to a missed case
    prevalence:
        -The prevalence of the disease in the population. Must be a float
    sensitivity:
        -The sensitivity level of the screening test. Must be a float
    specificity:
        -The specificity level of the screening test. Must be a float
    population:
        -The population size to set. Choose a larger value since this is only necessary for total calculations. Default is 10,000
    decimal:
        -amount of decimal points to display. Default value is 3
    """
    print('----------------------------------------------------------------------')
    print('''NOTE: When calculating costs, be sure to consult experts in health\npolicy or related fields.  
        Costs should encompass more than only monetary\ncosts, like relative costs (regret, disappointment, stigma, 
        disutility, etc.)''')
    if (sensitivity > 1) | (specificity > 1):
        raise ValueError('sensitivity/specificity/prevalence cannot be greater than 1')
    disease = population * prevalence
    disease_free = population - disease

    # TEST: no positives
    nt_cost = disease * cost_miss_case
    pc_nt_cost = nt_cost / population

    # TEST: all postives
    t_cost = disease_free * cost_false_pos
    pc_t_cost = t_cost / population

    # TEST: criteria
    cost_b = disease - (disease * sensitivity)
    cost_c = disease_free - (disease_free * specificity)
    ct_cost = (cost_miss_case * cost_b) + (cost_false_pos * cost_c)
    pc_ct_cost = ct_cost / population

    # Present results
    print('----------------------------------------------------------------------')
    print('Treat everyone as Test-Negative')
    print('Total relative cost:\t\t', round(nt_cost, decimal))
    print('Per Capita relative cost:\t', round(pc_nt_cost, decimal))
    print('----------------------------------------------------------------------')
    print('Treat everyone as Test-Positive')
    print('Total relative cost:\t\t', round(t_cost, decimal))
    print('Per Capita relative cost:\t', round(pc_t_cost, decimal))
    print('----------------------------------------------------------------------')
    print('Treating by Screening Test')
    print('Total relative cost:\t\t', round(ct_cost, decimal))
    print('Per Capita relative cost:\t', round(pc_ct_cost, decimal))
    print('----------------------------------------------------------------------')
    if pc_ct_cost > pc_nt_cost:
        print('Screening program is more costly than treating everyone as a test-negative')
    if pc_nt_cost > pc_ct_cost > pc_t_cost:
        print('Screening program is cost efficient')
    if (pc_t_cost < pc_ct_cost) and (pc_t_cost < pc_nt_cost):
        print('Treating everyone as test-positive is least costly')
    print('----------------------------------------------------------------------\n')



"""
_____________________________________________________________________________________________
|                                                                                           |
|       zEpid package: making epidemiology with Python e-z                                  |
|                                                                                           |
| This package is to provide easy to use tools to aid in epidemiological analyses. A full   |
| applied guide to the entire package is available at:                                      |
|       https://github.com/pzivich/Python-for-Epidemiologists                               |
|                                                                                           |
| Current contents include:                                                                 |
|       -Basics tools for pandas dataframes and summary data                                |
|       -Useful graphics                                                                    |
|       -Causal estimation methods                                                          |
|       -Sensitivity analysis tools                                                         |
|___________________________________________________________________________________________|

CONTENTS

Basic calculations (both for summary data and pandas.DataFrame):
    -Risk Ratio, Risk Difference, Number Needed to Treat, Odds Ratio, Incidence Rate Ratio,
     Incidence Rate Difference, Attributable Community Risk, Population Attributable Fraction,
     Interaction Contrast, Interaction Contrast Ratio, Sensitivity, Specificity, Splines

Causal methods:
    -Inverse Probability of Treatment Weights, Inverse Probability of Censoring Weights,
     Inverse Probability of Missing Weights, Time-Fixed G-Formula, Time-Fixed Double Robust
     Estimator

Graphics:
    -Forest Plot, Receiver Operator Curve, Functional Form Plot, P-value Distribution Plot,
     Spaghetti Plot, Dynamic Risk Plot

Sensitivity Analyses:
    -Monte Carlo correction for Risk Ratio, Trapezoidal distribution

Example)
>>>import zepid as ze
>>>ze.load_sample_data(timevary=False)
>>>ze.RiskDiff(df, exposure='art', outcome='dead') #Example of Risk Ratio Calculation

See http://zepid.readthedocs.io/en/latest/ for a full guide through all the package features
"""
from .base import (RiskRatio, RiskDiff, NNT, OddsRatio, IncRateRatio, IncRateDiff, ACR, PAF, IC, ICR,
                   Sensitivity, Specificity, Diagnostics, spline, Table1)
from .datasets import load_sample_data

import zepid.calc
import zepid.graphics
import zepid.causal.gformula
import zepid.causal.ipw
import zepid.causal.doublyrobust
import zepid.sensitivity_analysis

from .version import __version__



"""
Contains useful graphic generators. Currently, effect measure plots and functional form assessment plots
are implemented. Uses matplotlib to generate graphics. Future inclusions include forest plots

Contents:
Functional form assessment- func_form_plot()
Forest plot/ effect measure plot- EffectMeasurePlot()
P-value distribution plot- pvalue_plot()
Spaghetti plot- spaghetti_plot()
Receiver-Operator Curve- roc()
Dynamic risk plot- dynamic_risk_plot()
"""

import numpy as np
import pandas as pd
from scipy.stats import norm
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.genmod.families import links
from statsmodels.nonparametric.smoothers_lowess import lowess
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.ticker as mticker


class EffectMeasurePlot:
    """Used to generate effect measure plots. effectmeasure plot accepts four list type objects.
    effectmeasure_plot is initialized with the associated names for each line, the point estimate,
    the lower confidence limit, and the upper confidence limit.

    Plots will resemble the following form:

        _____________________________________________      Measure     % CI
        |                                           |
    1   |        --------o-------                   |       x        n, 2n
        |                                           |
    2   |                   ----o----               |       w        m, 2m
        |                                           |
        |___________________________________________|
        #           #           #           #



    The following functions (and their purposes) live within effectmeasure_plot

    labels(**kwargs)
        Used to change the labels in the plot, as well as the center and scale. Inputs are
        keyword arguments
        KEYWORDS:
            -effectmeasure  + changes the effect measure label
            -conf_int       + changes the confidence interval label
            -scale          + changes the scale to either log or linear
            -center         + changes the reference line for the center

    colors(**kwargs)
        Used to change the color of points and lines. Also can change the shape of points.
        Valid colors and shapes for matplotlib are required. Inputs are keyword arguments
        KEYWORDS:
            -errorbarcolor  + changes the error bar colors
            -linecolor      + changes the color of the reference line
            -pointcolor     + changes the color of the points
            -pointshape     + changes the shape of points

    plot(t_adjuster=0.01,decimal=3,size=3)
        Generates the effect measure plot of the input lists according to the pre-specified
        colors, shapes, and labels of the class object
        Arguments:
            -t_adjuster     + used to refine alignment of the table with the line graphs.
                              When generate plots, trial and error for this value are usually
                              necessary
            -decimal        + number of decimal places to display in the table
            -size           + size of the plot to generate


    Example)
    >>>lab = ['One','Two'] #generating lists of data to plot
    >>>emm = [1.01,1.31]
    >>>lcl = ['0.90',1.01]
    >>>ucl = [1.11,1.53]
    >>>
    >>>x = zepid.graphics.effectmeasure_plot(lab,emm,lcl,ucl) #initializing effectmeasure_plot with the above lists
    >>>x.labels(effectmeasure='RR') #changing the table label to 'RR'
    >>>x.colors(pointcolor='r') #changing the point colors to red
    >>>x.plot(t_adjuster=0.13) #generating the effect measure plot
    """

    def __init__(self, label, effect_measure, lcl, ucl):
        """Initializes effectmeasure_plot with desired data to plot. All lists should be the same
        length. If a blank space is desired in the plot, add an empty character object (' ') to
        each list at the desired point.

        Inputs:

        label
            -list of labels to use for y-axis
        effect_measure
            -list of numbers for point estimates to plot. If point estimate has trailing zeroes,
             input as a character object rather than a float
        lcl
            -list of numbers for upper confidence limits to plot. If point estimate has trailing
             zeroes, input as a character object rather than a float
        ucl
            -list of numbers for upper confidence limits to plot. If point estimate has
             trailing zeroes, input as a character object rather than a float
        """
        self.df = pd.DataFrame()
        self.df['study'] = label
        self.df['OR'] = effect_measure
        self.df['LCL'] = lcl
        self.df['UCL'] = ucl
        self.df['OR2'] = self.df['OR'].astype(str).astype(float)
        if (all(isinstance(item, float) for item in lcl)) & (all(isinstance(item, float) for item in effect_measure)):
            self.df['LCL_dif'] = self.df['OR'] - self.df['LCL']
        else:
            self.df['LCL_dif'] = (pd.to_numeric(self.df['OR'])) - (pd.to_numeric(self.df['LCL']))
        if (all(isinstance(item, float) for item in ucl)) & (all(isinstance(item, float) for item in effect_measure)):
            self.df['UCL_dif'] = self.df['UCL'] - self.df['OR']
        else:
            self.df['UCL_dif'] = (pd.to_numeric(self.df['UCL'])) - (pd.to_numeric(self.df['OR']))
        self.em = 'OR'
        self.ci = '95% CI'
        self.scale = 'linear'
        self.center = 1
        self.errc = 'dimgrey'
        self.shape = 'd'
        self.pc = 'k'
        self.linec = 'gray'

    def labels(self, **kwargs):
        """Function to change the labels of the outputted table. Additionally, the scale and reference
        value can be changed.

        Accepts the following keyword arguments:

        effectmeasure
            -changes the effect measure label
        conf_int
            -changes the confidence interval label
        scale
            -changes the scale to either log or linear
        center
            -changes the reference line for the center
        """
        if 'effectmeasure' in kwargs:
            self.em = kwargs['effectmeasure']
        if 'ci' in kwargs:
            self.ci = kwargs['conf_int']
        if 'scale' in kwargs:
            self.scale = kwargs['scale']
        if 'center' in kwargs:
            self.center = kwargs['center']

    def colors(self, **kwargs):
        """Function to change colors and shapes.

        Accepts the following keyword arguments:

        errorbarcolor
            -changes the error bar colors
        linecolor
            -changes the color of the reference line
        pointcolor
            -changes the color of the points
        pointshape
            -changes the shape of points
        """
        if 'errorbarcolor' in kwargs:
            self.errc = kwargs['errorbarcolor']
        if 'pointshape' in kwargs:
            self.shape = kwargs['pointshape']
        if 'linecolor' in kwargs:
            self.linec = kwargs['linecolor']
        if 'pointcolor' in kwargs:
            self.pc = kwargs['pointcolor']

    def plot(self, figsize=(3, 3), t_adjuster=0.01, decimal=3, size=3, max_value=None, min_value=None):
        """Generates the matplotlib effect measure plot with the default or specified attributes.
        The following variables can be used to further fine-tune the effect measure plot

        t_adjuster
            -used to refine alignment of the table with the line graphs. When generate plots, trial
             and error for this value are usually necessary. I haven't come up with an algorithm to
             determine this yet...
        decimal
            -number of decimal places to display in the table
        size
            -size of the plot to generate
        max_value
            -maximum value of x-axis scale. Default is None, which automatically determines max value
        min_value
            -minimum value of x-axis scale. Default is None, which automatically determines min value
        """
        tval = []
        ytick = []
        for i in range(len(self.df)):
            if (np.isnan(self.df['OR2'][i]) == False):
                if ((isinstance(self.df['OR'][i], float)) & (isinstance(self.df['LCL'][i], float)) & (
                isinstance(self.df['UCL'][i], float))):
                    tval.append([round(self.df['OR2'][i], decimal), (
                                '(' + str(round(self.df['LCL'][i], decimal)) + ', ' + str(
                            round(self.df['UCL'][i], decimal)) + ')')])
                else:
                    tval.append(
                        [self.df['OR'][i], ('(' + str(self.df['LCL'][i]) + ', ' + str(self.df['UCL'][i]) + ')')])
                ytick.append(i)
            else:
                tval.append([' ', ' '])
                ytick.append(i)
        if max_value is None:
            if pd.to_numeric(self.df['UCL']).max() < 1:
                maxi = round(((pd.to_numeric(self.df['UCL'])).max() + 0.05),
                             2)  # setting x-axis maximum for UCL less than 1
            if (pd.to_numeric(self.df['UCL']).max() < 9) and (pd.to_numeric(self.df['UCL']).max() >= 1):
                maxi = round(((pd.to_numeric(self.df['UCL'])).max() + 1),
                             0)  # setting x-axis maximum for UCL less than 10
            if pd.to_numeric(self.df['UCL']).max() > 9:
                maxi = round(((pd.to_numeric(self.df['UCL'])).max() + 10),
                             0)  # setting x-axis maximum for UCL less than 100
        else:
            maxi = max_value
        if min_value is None:
            if pd.to_numeric(self.df['LCL']).min() > 0:
                mini = round(((pd.to_numeric(self.df['LCL'])).min() - 0.1), 1)  # setting x-axis minimum
            if pd.to_numeric(self.df['LCL']).min() < 0:
                mini = round(((pd.to_numeric(self.df['LCL'])).min() - 0.05), 2)  # setting x-axis minimum
        else:
            mini = min_value
        plt.figure(figsize=figsize)  # blank figure
        gspec = gridspec.GridSpec(1, 6)  # sets up grid
        plot = plt.subplot(gspec[0, 0:4])  # plot of data
        tabl = plt.subplot(gspec[0, 4:])  # table of OR & CI
        plot.set_ylim(-1, (len(self.df)))  # spacing out y-axis properly
        if self.scale == 'log':
            try:
                plot.set_xscale('log')
            except:
                raise ValueError('For the log scale, all values must be positive')
        plot.axvline(self.center, color=self.linec, zorder=1)
        plot.errorbar(self.df.OR2, self.df.index, xerr=[self.df.LCL_dif, self.df.UCL_dif], marker='None', zorder=2,
                      ecolor=self.errc, elinewidth=(size / size), linewidth=0)
        plot.scatter(self.df.OR2, self.df.index, c=self.pc, s=(size * 25), marker=self.shape, zorder=3,
                     edgecolors='None')
        plot.xaxis.set_ticks_position('bottom')
        plot.yaxis.set_ticks_position('left')
        plot.get_xaxis().set_major_formatter(matplotlib.ticker.ScalarFormatter())
        plot.get_xaxis().set_minor_formatter(matplotlib.ticker.NullFormatter())
        plot.set_yticks(ytick)
        plot.set_xlim([mini, maxi])
        plot.set_xticks([mini, self.center, maxi])
        plot.set_xticklabels([mini, self.center, maxi])
        plot.set_yticklabels(self.df.study)
        plot.yaxis.set_ticks_position('none')
        plot.invert_yaxis()  # invert y-axis to align values properly with table
        tb = tabl.table(cellText=tval, cellLoc='center', loc='right', colLabels=[self.em, self.ci],
                        bbox=[0, t_adjuster, 1, 1])
        tabl.axis('off')
        tb.auto_set_font_size(False)
        tb.set_fontsize(12)
        for key, cell in tb.get_celld().items():
            cell.set_linewidth(0)
        return plot


def func_form_plot(df, outcome, var, f_form=None, outcome_type='binary', link_dist=None, ylims=None, loess_value=0.25,
                   legend=True, model_results=True, loess=True, points=False, discrete=False):
    """Creates a LOESS plot to aid in functional form assessment for continuous variables.
    Plots can be created for binary and continuous outcomes. Default options are set to create
    a functional form plot for a binary outcome. To convert to a continuous outcome,
    outcome_type needs to be changed, in addition to the link_dist

    Returns a matplotlib graph with a LOESS line (dashed red-line), regression line (sold blue-line),
    and confidence interval (shaded blue)

    df:
        -dataframe that contains the variables of interest
    outcome:
        -Column name of the outcome variable of interest
    var:
        -Column name of the variable of interest for the functional form assessment
    f_form:
        -Regression equation of the functional form to assess. Default is None, which will produce
         a linear functional form. Input the regression equation as the variables of interest, separated
         by +. Example) 'var + var_sq'
    outcome_type:
        -Variable type of the outcome variable. Currently, only binary and continuous variables are
         supported. Default is binary
    link_dist:
        -Link and distribution for the GLM regression equation. Change this to any valid link and
        distributions supported by statsmodels. Default is None, which conducts logistic regression
    ylims:
        -List object of length 2 that holds the upper and lower limits of the y-axis. Y-axis limits should be
        specified when comparing multiple graphs. These need to be user-specified since the results between
        models and datasets can be so variable. Default is None, which returns the matplotlib y-axis of best fit.
    loess_value:
        -Fraction of observations to use to fit the LOESS curve. This will need to be changed iteratively
         to determine which percent works best for the data. Default is 0.5
    legend:
        -Turn the legend on or off. Default is True, displaying the legend in the graph
    model_results:
        -Whether to produce the model results. Default is True, which provides model results
    loess:
        -Whether to plot the LOESS curve along with the functional form. Default is True
    points:
        -Whether to plot the data points, where size is relative to the number of observations. Default is False
    discrete:
        -If your data is truly continuous, leave setting to bin the dat. Will automatically bin observations into categories
         for generation of the LOESS curve. If you data is discrete, you can set this to True to use your actual values.
         If you get a perfect SeparationError from statsmodels, it means you might have to reshift your categories.

    Example)
    >>>data['var1_sq'] = data['var1']**2
    >>>zepid.graphics.func_form_plot(df=data,outcome='D',var='var1',f_form='var1 + var1_sq')
    """
    # Copying out the dataframe to a new object we will manipulate a bit
    rf = df.copy()
    rf = rf.dropna(subset=[var, outcome]).sort_values(by=[var, outcome]).reset_index()
    print('Warning: missing observations of model variables are dropped')
    print(int(df.shape[0] - rf.shape[0]), ' observations were dropped from the functional form assessment')

    # Functional form for the model
    if f_form is None:
        f_form = var
    else:
        pass

        # Generating Models
    if link_dist is None:
        link_dist = sm.families.family.Binomial(sm.families.links.logit)
    else:
        pass

    # Generating LOESS or points if requested
    if loess or points:
        if outcome_type == 'binary':
            if discrete is False:
                # Binning continuous variable into categories to get "General" functional form
                categories = int((np.max(rf[var]) - np.min(rf[var])) / 5)
                print('''A total of ''' + str(categories) + ''' categories were created. If you would like to influence the 
                       number of categories the spline is fit to, do the following\n\tIncrease: multiply by a constant 
                       >1\n\tDecrease: multiply by a contast <1 and >0''')
                rf['vbin'] = pd.qcut(rf[var], q=categories, duplicates='drop').cat.codes
                djm = smf.glm(outcome + '~ C(vbin)', rf, family=link_dist).fit()
            else:
                djm = smf.glm(outcome + '~ C(' + var + ')', rf, family=link_dist).fit()
            djf = djm.get_prediction(rf).summary_frame()
            dj = pd.concat([rf, djf], axis=1)
            dj.sort_values(var, inplace=True)
            if points:
                pf = dj.groupby(by=[var, 'mean']).count().reset_index()
            if loess:
                yl = lowess(list(dj['mean']), list(dj[var]), frac=loess_value)
                lowess_x = list(zip(*yl))[0]
                lowess_y = list(zip(*yl))[1]
        elif outcome_type == 'continuous':
            if points:
                pf = rf.groupby(by=[var, outcome]).count().reset_index()
            if loess:
                yl = lowess(list(rf[outcome]), list(rf[var]), frac=loess_value)
                lowess_x = list(zip(*yl))[0]
                lowess_y = list(zip(*yl))[1]
        else:
            raise ValueError('Functional form assessment only supports binary or continuous outcomes currently')

    # Functional form model fitting
    ffm = smf.glm(outcome + ' ~ ' + f_form, rf, family=link_dist).fit()
    if model_results is True:
        print(ffm.summary())
        print('AIC: ', ffm.aic)
        print('BIC: ', ffm.bic)
    fff = ffm.get_prediction(rf).summary_frame()
    ff = pd.concat([rf, fff], axis=1)
    ff.sort_values(var, inplace=True)

    # Generating plot for functional form
    ax = plt.gca()
    if points is True:
        if outcome_type == 'continuous':
            ax.scatter(pf[var], pf[outcome], s=[100 * (n / np.max(pf[var])) for n in pf[var]],
                       color='gray', label='Data point')
        else:
            ax.scatter(pf[var], pf['mean'], s=[100 * (n / np.max(pf[var])) for n in pf[var]],
                       color='gray', label='Data point')
    ax.fill_between(ff[var], ff['mean_ci_upper'], ff['mean_ci_lower'], alpha=0.1,
                    color='blue', label='95% CI')
    ax.plot(ff[var], ff['mean'], '-', color='blue', label='Regression')
    if loess is True:
        ax.plot(lowess_x, lowess_y, '--', color='red', linewidth=1, label='LOESS')
    ax.set_xlabel(var)
    if outcome_type is 'binary':
        ax.set_ylabel('Probability')
    else:
        ax.set_ylabel(outcome)
    if legend is True:
        ax.legend()
    ax.set_ylim(ylims)
    return ax


def pvalue_plot(point, sd, color='b', fill=True, null=0, alpha=None):
    """Creates a plot of the p-value distribution based on a point estimate and standard deviation.
    I find this plot to be useful to explain p-values and how much evidence weight you have in a
    specific value. I think it is useful to explain what exactly a p-value tells you. Note that this
    plot only works for measures on a linear scale (i.e. it will plot exp(log(RR)) incorrectly). It also
    helps to understand what exactly confidence intervals are telling you. These  plots are based on
    Rothman Epidemiology 2nd Edition pg 152-153 and explained more fully within.

    Returns matplotlib axes object

    point:
        -point estimate. Must be on a linear scale (RD / log(RR))
    sd:
        -standard error of the estimate. Must for linear scale (SE(RD) / SE(log(RR)))
    color:
        -change color of p-value plot
    fill:
        -Whether to fill the curve under the p-value distribution. Setting to False prevents fill
    null:
        -The main value to compare to. The default is zero

    Example)
    >>>zepid.graphics.pvalue_plot(point=-0.1,sd=0.061,alpha=0.025)
    """
    if point <= null:
        lower = (point - 3 * sd)
        if (point + 3 * sd) < 0:
            upper = point + 3 * sd
        else:
            upper = null + 3 * sd
    if point > null:
        upper = (point + 3 * sd)
        if (point - 3 * sd) > 0:
            lower = null - 3 * sd
        else:
            lower = point - 3 * sd
    ax = plt.gca()
    x1 = np.linspace(lower, point, 100)
    x2 = np.linspace(point, upper, 100)
    ax.plot(x2, 2 * (1 - norm.cdf(x2, loc=point, scale=sd)), c=color)
    ax.plot(x1, 2 * norm.cdf(x1, loc=point, scale=sd), c=color)
    if fill == True:
        ax.fill_between(x2, 2 * (1 - norm.cdf(x2, loc=point, scale=sd)), color=color, alpha=0.2)
        ax.fill_between(x1, 2 * norm.cdf(x1, loc=point, scale=sd), color=color, alpha=0.2)
    ax.vlines(null, 0, 1, colors='k')
    ax.set_xlim([lower, upper])
    ax.set_ylim([0, 1])
    ax.set_ylabel('P-value')
    if alpha is not None:
        ax.hlines(alpha, lower, upper)
    return ax


def spaghetti_plot(df, idvar, variable, time):
    """Create a spaghetti plot by an ID variable. A spaghetti plot can be useful for visualizing
    trends or looking at longitudinal data patterns for individuals all at once.

    Returns matplotlib axes

    df:
        -pandas dataframe containing variables of interest
    idvar:
        -ID variable for observations. This should indicate the group or individual followed over
         the time variable
    variable:
        -Variable of interest to see how it varies over time
    time:
        -Time or other variable in which the variable variation occurs

    Example)
    >>>zepid.graphics.spaghetti_plot(df,idvar='pid',variable='v',time='t')
    """
    ax = plt.gca()
    for i in df[idvar].unique():
        s = df.loc[df[idvar] == i].copy()
        s.sort_values(time, ascending=False)
        ax.plot(s[time], s[variable])
    ax.set_xlabel(time)
    ax.set_ylabel(variable)
    return ax


def roc(df, true, threshold, youden_index=True):
    """Generate a Receiver Operator Curve from true values and predicted probabilities.

    Returns matplotlib axes

    df:
        -pandas dataframe containing variables of interest
    true:
        -the true designation of the outcome (1, 0)
    threshold:
        -predicted probabilities for the outcome
    youden_index:
        -Whether to calculate Youden's index. Youden's index maximizes both sensitivity and specificity.
         The formula finds the maximum of (sensitivity + specificity - 1)
    """
    sens = []
    fpr = []
    thresh = []
    tf = df[[threshold, true]].copy()
    if tf.isnull().values.sum() != 0:
        raise ValueError('ROC curve cannot handle missing data for probability or true values')

    # Getting all possible cutpoints
    values = (list(np.unique(tf[threshold])))
    values = [float(np.min(tf[threshold]) - 0.001)] + values + [float(np.max(tf[threshold]) + 0.001)]
    # Going through all the cutpoints and calculating Sensitivity and 1-Specificity
    for v in list(reversed(values)):
        thresh.append(v)
        prediction = np.where(tf[threshold] >= v, 1, 0)
        se = prediction[tf[true] == 1].mean()
        sens.append(se)
        sp = prediction[tf[true] == 0].mean()
        fpr.append(sp)

    # If requested, calculate Youden's Index
    if youden_index is True:
        spec = [1 - i for i in fpr]
        youdens = []
        for i, j in zip(sens, spec):
            youdens.append(i + j - 1)
        ind = np.argmax(youdens)
        print('----------------------------------------------------------------------')
        print("Youden's Index: ", thresh[ind])
        print("Predictive values at Youden's Index")
        print("\tSensitivity: ", sens[ind])
        print("\tSpecificity: ", spec[ind])
        print('----------------------------------------------------------------------')

    # Creating ROC plot
    ax = plt.gca()
    ax.plot(fpr, sens, color='blue')
    ax.plot([0, 1], [0, 1], color='gray', linestyle='--')
    if youden_index is True:
        ax.text(0.65, 0.35, "Youden's Index:\n      " + str(round(thresh[ind], 5)))
    ax.set_xlim([-0.01, 1.01])
    ax.set_ylim([-0.01, 1.01])
    ax.set_ylabel('Sensitivity')
    ax.set_xlabel('1 -Specificity')
    return ax


def dyanmic_risk_plot(risk_exposed, risk_unexposed, measure='RD', loess=True, loess_value=0.25, point_color='darkblue',
                      line_color='b', scale='linear'):
    """Creates a plot of risk measures over time. See Cole et al. "Estimation of standardized risk difference and ratio
    in a competing risks framework: application to injection drug use and progression to AIDS after initiation of
    antiretroviral therapy." Am J Epidemiol. 2015 for an example of this plot

    risk_exposed:
        -pandas Series with the probability of the outcome among the exposed group. Index by 'timeline'
         where 'timeline' is the time. If you directly output the 1 - survival_function_ from
         lifelines.KaplanMeierFitter(), this should create a valid input
    risk_unexposed:
        -pandas Series with the probability of the outcome among the exposed group. Index by 'timeline'
         where 'timeline' is the time
    measure:
        -whether to generate the risk difference (RD) or risk ratio (RR). Default is 'RD'
    loess:
        -whether to generate LOESS curve fit to the calculated points. Default is True
    loess_value:
        -fraction of values to fit LOESS curve to. Default is 0.25
    point_color:
        -color of the points
    line_color:
        -color of the LOESS line generated and plotted
    scale:
        -change the y-axis scale. Options are 'linear' (default), 'log', 'log-transform'. 'log' and 'log-transform'
         is only a valid option for Risk Ratio plots
    """
    re = risk_exposed.drop_duplicates(keep='first').iloc[:, 0].rename('exposed').reset_index()
    ru = risk_unexposed.drop_duplicates(keep='first').iloc[:, 0].rename('unexposed').reset_index()
    re.timeline = np.round(re.timeline * 100000).astype(int) # This avoids a merge issue on floats
    ru.timeline = np.round(ru.timeline * 100000).astype(int)
    r = pd.merge(re, ru, how='outer', left_on='timeline', right_on='timeline').sort_values(by='timeline')
    r.timeline /= 100000
    r.ffill(inplace=True)
    if measure == 'RD':
        r['m'] = r['exposed'] - r['unexposed']
    elif measure == 'RR':
        r['m'] = r['exposed'] / r['unexposed']
        if scale == 'log-transform':
            r['m'] = np.log(r['m'])
    else:
        raise ValueError('Only "RD" and "RR" are currently supported')

    # Generating the plot
    ax = plt.gca()
    ax.plot(r['timeline'], r['m'], 'o', c=point_color)
    if loess is True:
        l = lowess(list(r['m']), list(r['timeline']), frac=loess_value)
        lowess_x = list(zip(*l))[0]
        lowess_y = list(zip(*l))[1]
        ax.plot(lowess_x, lowess_y, '-', c=line_color, linewidth=4)
    if measure == 'RD':
        ax.hlines(0, 0, np.max(r['timeline'] + 0.5), linewidth=1.5)
        ax.set_ylabel('Risk Difference')
    if measure == 'RR':
        if scale == 'log-transform':
            ax.hlines(0, 0, np.max(r['timeline'] + 0.5), linewidth=1.5)
            ax.set_ylabel('ln(Risk Ratio)')
        elif scale == 'log':
            ax.set_ylabel('Risk Ratio')
            ax.set_yscale('log')
            ax.yaxis.set_major_formatter(mticker.ScalarFormatter())
            ax.yaxis.get_major_formatter().set_scientific(False)
            ax.yaxis.get_major_formatter().set_useOffset(False)
            ax.hlines(1, 0, np.max(r['timeline'] + 0.5), linewidth=1.5)
        else:
            ax.hlines(1, 0, np.max(r['timeline'] + 0.5), linewidth=1.5)
    ax.set_xlabel('Time')
    ax.set_xlim([0, np.max(r['timeline']) + 0.5])
    return ax



import warnings
import math
import numpy as np
import pandas as pd
from scipy.stats import norm
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.genmod.families import links
from tabulate import tabulate
from zepid.calc.utils import rr, rd, nnt, oddsratio, ird, irr, acr, paf, sensitivity, specificity


def RiskRatio(df, exposure, outcome, reference=0, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """Estimate of Relative Risk with a (1-alpha)*100% Confidence interval. Missing data is ignored by
    this function.

    WARNING: Exposure & Outcome must be coded as (1: yes, 0:no). Only works for binary exposures and outcomes

    df:
        -pandas dataframe containing variables of interest
    exposure:
        -column name of exposure variable. Must be coded as binary (0,1) where 1 is exposed
    outcome:
        -column name of outcome variable. Must be coded as binary (0,1) where 1 is the outcome of interest
    reference:
        -reference category for comparisons
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% confidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the RR as a object. Default is False

    Example)
    >>>zepid.RiskRatio(df=data,exposure='X',outcome='D')
    """
    c = df.loc[(df[exposure] == reference) & (df[outcome] == 1)].shape[0]
    d = df.loc[(df[exposure] == reference) & (df[outcome] == 0)].shape[0]
    vals = set(df[exposure].dropna().unique())
    vals.remove(reference)
    for i in vals:
        print('======================================================================')
        print('Comparing ' + exposure + '=' + str(i) + ' to ' + exposure + '=' + str(reference))
        print('======================================================================')
        a = df.loc[(df[exposure] == i) & (df[outcome] == 1)].shape[0]
        b = df.loc[(df[exposure] == i) & (df[outcome] == 0)].shape[0]
        rr(a=a, b=b, c=c, d=d, alpha=alpha, decimal=decimal, print_result=print_result, return_result=return_result)


def RiskDiff(df, exposure, outcome, reference=0, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """Estimate of Risk Difference with a (1-alpha)*100% Confidence interval. Missing data is ignored by this
    function.

    WARNING: Exposure & Outcome must be coded as 1 and 0 for this to work properly (1: yes, 0:no)

    df:
        -pandas dataframe containing the variables of interest
    exposure:
        -column name of exposure variable. Must be coded as binary (0,1) where 1 is exposed
    outcome:
        -column name of outcome variable. Must be coded as binary (0,1) where 1 is the outcome of interest
    reference:
        -reference category for comparisons
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True, which prints the results
    return_result:
        -Whether to return the RD as a object. Default is False

    Example)
    >>>zepid.RiskDiff(df=data,exposure='X',outcome='D')
    """
    c = df.loc[(df[exposure] == reference) & (df[outcome] == 1)].shape[0]
    d = df.loc[(df[exposure] == reference) & (df[outcome] == 0)].shape[0]
    vals = set(df[exposure].dropna().unique())
    vals.remove(reference)
    for i in vals:
        print('======================================================================')
        print('Comparing ' + exposure + '=' + str(i) + ' to ' + exposure + '=' + str(reference))
        print('======================================================================')
        a = df.loc[(df[exposure] == i) & (df[outcome] == 1)].shape[0]
        b = df.loc[(df[exposure] == i) & (df[outcome] == 0)].shape[0]
        rd(a=a, b=b, c=c, d=d, alpha=alpha, decimal=decimal, print_result=print_result, return_result=return_result)


def NNT(df, exposure, outcome, reference=0, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """Estimates of Number Needed to Treat. NNT (1-alpha)*100% confidence interval presentation is based on
    Altman, DG (BMJ 1998). Missing data is ignored by this function.

    WARNING: Exposure & Outcome must be coded as 1 and 0 for this to work properly (1: yes, 0:no).

    df:
        -pandas dataframe containing the variables of interest
    exposure:
        -column name of exposure variable. Must be coded as binary (0,1) where 1 is exposed
    outcome:
        -column name of outcome variable. Must be coded as binary (0,1) where 1 is the outcome of interest
    reference:
        -reference category for comparisons
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the NNT as a object. Default is False

        Example)
    >>>zepid.NNT(df=data,exposure='X',outcome='D')
    """
    c = df.loc[(df[exposure] == reference) & (df[outcome] == 1)].shape[0]
    d = df.loc[(df[exposure] == reference) & (df[outcome] == 0)].shape[0]
    vals = set(df[exposure].dropna().unique())
    vals.remove(reference)
    for i in vals:
        print('======================================================================')
        print('Comparing ' + exposure + '=' + str(i) + ' to ' + exposure + '=' + str(reference))
        print('======================================================================')
        a = df.loc[(df[exposure] == i) & (df[outcome] == 1)].shape[0]
        b = df.loc[(df[exposure] == i) & (df[outcome] == 0)].shape[0]
        nnt(a=a, b=b, c=c, d=d, alpha=alpha, decimal=decimal, print_result=print_result, return_result=return_result)


def OddsRatio(df, exposure, outcome, reference=0, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """Estimates of Odds Ratio with a (1-alpha)*100% Confidence interval. Missing data is ignored by this function.

    WARNING: Exposure & Outcome must be coded as 1 and 0 for this to work properly (1: yes, 0:no).

    df:
        -pandas dataframe containing the variables of interest
    exposure:
        -column name of exposure variable. Must be coded as binary (0,1) where 1 is exposed
    outcome:
        -column name of outcome variable. Must be coded as binary (0,1) where 1 is the outcome of interest
    reference:
        -reference category for comparisons
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the RR as a object. Default is False

    Example)
    >>>zepid.OddsRatio(df=data,exposure='X',outcome='D')
    """
    c = df.loc[(df[exposure] == reference) & (df[outcome] == 1)].shape[0]
    d = df.loc[(df[exposure] == reference) & (df[outcome] == 0)].shape[0]
    vals = set(df[exposure].dropna().unique())
    vals.remove(reference)
    for i in vals:
        print('======================================================================')
        print('Comparing ' + exposure + '=' + str(i) + ' to ' + exposure + '=' + str(reference))
        print('======================================================================')
        a = df.loc[(df[exposure] == i) & (df[outcome] == 1)].shape[0]
        b = df.loc[(df[exposure] == i) & (df[outcome] == 0)].shape[0]
        oddsratio(a=a, b=b, c=c, d=d, alpha=alpha, decimal=decimal, print_result=print_result,
                  return_result=return_result)


def IncRateRatio(df, exposure, outcome, time, reference=0, alpha=0.05, decimal=3, print_result=True,
                 return_result=False):
    """Produces the estimate of the Incidence Rate Ratio with a (1-*alpha)*100% Confidence Interval.
    Missing data is ignored by this function.

    WARNING: Exposure & Outcome must be coded as 1 and 0 (1: yes, 0: no).

    df:
        -pandas dataframe containing the variables of interest
    exposure:
        -column name of exposure variable. Must be coded as binary (0,1) where 1 is exposed
    outcome:
        -column name of outcome variable. Must be coded as binary (0,1) where 1 is the outcome of interest
    time:
        -column name of person-time contributed by each individual. Must all be greater than 0
    reference:
        -reference category for comparisons
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the RR as a object. Default is False
    """
    c = df.loc[(df[exposure] == reference) & (df[outcome] == 1)].shape[0]
    time_c = df.loc[df[exposure] == reference][time].sum()
    vals = set(df[exposure].dropna().unique())
    vals.remove(reference)
    for i in vals:
        print('======================================================================')
        print('Comparing ' + exposure + '=' + str(i) + ' to ' + exposure + '=' + str(reference))
        print('======================================================================')
        a = df.loc[(df[exposure] == i) & (df[outcome] == 1)].shape[0]
        time_a = df.loc[df[exposure] == i][time].sum()
        irr(a=a, c=c, t1=time_a, t2=time_c, alpha=alpha, decimal=decimal, print_result=print_result,
            return_result=return_result)


def IncRateDiff(df, exposure, outcome, time, reference=0, alpha=0.05, decimal=3, print_result=True,
                return_result=False):
    """Produces the estimate of the Incidence Rate Difference with a (1-alpha)*100% confidence interval.
    Missing data is ignored by this function.

    WARNING: Exposure & Outcome must be coded as 1 and 0 (1: yes, 0: no)

    exposure:
        -column name of exposure variable. Must be coded as binary (0,1) where 1 is exposed
    outcome:
        -column name of outcome variable. Must be coded as binary (0,1) where 1 is the outcome of interest
    time:
        -column name of person-time contributed by individual. Must be greater than 0
    reference:
        -reference category for comparisons
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the RR as a object. Default is False
    """
    c = df.loc[(df[exposure] == reference) & (df[outcome] == 1)].shape[0]
    time_c = df.loc[df[exposure] == reference][time].sum()
    vals = set(df[exposure].dropna().unique())
    vals.remove(reference)
    for i in vals:
        print('======================================================================')
        print('Comparing ' + exposure + '=' + str(i) + ' to ' + exposure + '=' + str(reference))
        print('======================================================================')
        a = df.loc[(df[exposure] == i) & (df[outcome] == 1)].shape[0]
        time_a = df.loc[df[exposure] == i][time].sum()
        ird(a=a, c=c, t1=time_a, t2=time_c, alpha=alpha, decimal=decimal, print_result=print_result,
            return_result=return_result)


def ACR(df, exposure, outcome, decimal=3):
    """Produces the estimated Attributable Community Risk (ACR). ACR is also known as Population Attributable
    Risk. Since this is commonly confused with the population attributable fraction, the name ACR is used to
    clarify differences in the formulas. Missing data is ignored by this function.

    WARNING: Exposure & Outcome must be coded as 1 and 0 for this to work properly (1: yes, 0:no)

    df:
        -pandas dataframe containing the variables of interest
    exposure:
        -column name of exposure variable. Must be coded as binary (0,1) where 1 is exposed
    outcome:
        -column name of outcome variable. Must be coded as binary (0,1) where 1 is the outcome of interest
    decimal:
        -amount of decimal points to display. Default is 3
    """
    a = df.loc[(df[exposure] == 1) & (df[outcome] == 1)].shape[0]
    b = df.loc[(df[exposure] == 1) & (df[outcome] == 0)].shape[0]
    c = df.loc[(df[exposure] == 0) & (df[outcome] == 1)].shape[0]
    d = df.loc[(df[exposure] == 0) & (df[outcome] == 0)].shape[0]
    acr(a=a, b=b, c=c, d=d, decimal=decimal)


def PAF(df, exposure, outcome, decimal=3):
    """Produces the estimated Population Attributable Fraction. Missing data is ignored by this function.

    WARNING: Exposure & Outcome must be coded as 1 and 0 for this to work properly (1: yes, 0:no)

    exposure:
        -column name of exposure variable. Must be coded as binary (0,1) where 1 is exposed
    outcome:
        -column name of outcome variable. Must be coded as binary (0,1) where 1 is the outcome of interest
    decimal:
        -amount of decimal points to display. Default is 3
    """
    a = df.loc[(df[exposure] == 1) & (df[outcome] == 1)].shape[0]
    b = df.loc[(df[exposure] == 1) & (df[outcome] == 0)].shape[0]
    c = df.loc[(df[exposure] == 0) & (df[outcome] == 1)].shape[0]
    d = df.loc[(df[exposure] == 0) & (df[outcome] == 0)].shape[0]
    paf(a=a, b=b, c=c, d=d, decimal=decimal)


def IC(df, exposure, outcome, modifier, adjust=None, decimal=3):
    """Calculate the Interaction Contrast (IC) using a pandas dataframe and statsmodels to fit a linear
    binomial regression. Can ONLY be used for a 0,1 coded exposure and modifier (exposure = {0,1}, modifier = {0,1},
    outcome = {0,1}). Can handle adjustment for other confounders in the regression model. Prints the fit
    of the linear binomial regression, the IC, and the corresponding IC 95% confidence interval.

    NOTE: statsmodels may produce a domain error in some versions.

    df:
        -pandas dataframe containing variables of interest
    exposure:
        -column name of exposure variable. Must be coded as (0,1) where 1 is exposure
    outcome:
        -column name of outcome variable. Must be coded as (0,1) where 1 is outcome of interest
    modifier:
        -column name of modifier variable. Must be coded as (0,1) where 1 is modifier
    adjust:
        -string of other variables to adjust for, in correct statsmodels format. Default is None
        NOTE: variables can NOT be named {E1M0,E0M1,E1M1} since this function creates variables with those names.
              Answers will be incorrect
         Ex) '+ C1 + C2 + C3 + Z'
    decimal:
        -Number of decimals to display in result. Default is 3

    Example)
    >>>zepid.IC(df=data,exposure='X',outcome='D',modifier='Z',adjust='var1 + var2')
                     Generalized Linear Model Regression Results
    ==============================================================================
    Dep. Variable:                      D   No. Observations:                  210
    Model:                            GLM   Df Residuals:                      204
    Model Family:                Binomial   Df Model:                            5
    Link Function:               identity   Scale:                             1.0
    Method:                          IRLS   Log-Likelihood:                -97.450
    Date:                Thu, 03 May 2018   Deviance:                       194.90
    Time:                        18:46:13   Pearson chi2:                     198.
    No. Iterations:                    79
    ==============================================================================
                     coef    std err          z      P>|z|      [0.025      0.975]
    ------------------------------------------------------------------------------
    Intercept      0.6101      0.542      1.125      0.260      -0.453       1.673
    X              0.2049      0.056      3.665      0.000       0.095       0.314
    Z              0.1580      0.049      3.207      0.001       0.061       0.255
    E1M1          -0.2105      0.086     -2.447      0.014      -0.379      -0.042
    var1        7.544e-05    6.7e-05      1.125      0.260    -5.6e-05       0.000
    var2          -0.0248      0.022     -1.125      0.260      -0.068       0.018
    ==============================================================================

    ----------------------------------------------------------------------
    Interaction Contrast
    ----------------------------------------------------------------------

    IC:		-0.21047
    95% CI:		(-0.37908, -0.04186)
    ----------------------------------------------------------------------
    """
    df.loc[((df[exposure] == 1) & (df[modifier] == 1)), 'E1M1'] = 1
    df.loc[((df[exposure] != 1) | (df[modifier] != 1)), 'E1M1'] = 0
    df.loc[((df[exposure].isnull()) | (df[modifier].isnull())), 'E1M1'] = np.nan
    if adjust is None:
        eq = outcome + ' ~ ' + exposure + ' + ' + modifier + ' + E1M1'
    else:
        eq = outcome + ' ~ ' + exposure + ' + ' + modifier + ' + E1M1 + ' + adjust
    f = sm.families.family.Binomial(sm.families.links.identity)
    model = smf.glm(eq, df, family=f).fit()
    print(model.summary())
    ic = model.params['E1M1']
    lcl = model.conf_int().loc['E1M1'][0]
    ucl = model.conf_int().loc['E1M1'][1]
    print('\n----------------------------------------------------------------------')
    print('Interaction Contrast')
    print('----------------------------------------------------------------------')
    print('\nIC:\t\t' + str(round(ic, decimal)))
    print('95% CI:\t\t(' + str(round(lcl, decimal)) + ', ' + str(round(ucl, decimal)) + ')')
    print('----------------------------------------------------------------------')


def ICR(df, exposure, outcome, modifier, adjust=None, regression='log', ci='delta', b_sample=200, alpha=0.05,
        decimal=5):
    """Calculate the Interaction Contrast Ratio (ICR) using a pandas dataframe, and conducts either log binomial
    or logistic regression through statsmodels. Can ONLY be used for a 0,1 coded exposure and modifier (exposure = {0,1},
    modifier = {0,1}, outcome = {0,1}). Can handle missing data and adjustment for other confounders in the regression
    model. Prints the fit of the binomial regression, the ICR, and the corresponding ICR confidence interval. Confidence
    intervals can be generated using the delta method or bootstrap method

    NOTE: statsmodels may produce a domain error for log binomial models in some versions

    df:
        -pandas dataframe containing variables of interest
    exposure:
        -column name of exposure variable. Must be coded as (0,1) where 1 is exposure
    outcome:
        -column name of outcome variable. Must be coded as (0,1) where 1 is outcome of interest
    modifier:
        -column name of modifier variable. Must be coded as (0,1) where 1 is modifier
    adjust:
        -string of other variables to adjust for, in correct statsmodels format. Default is none
        NOTE: variables can NOT be named {E1M0,E0M1,E1M1} since this function creates variables with those names.
              Answers will be incorrect
         Ex) '+ C1 + C2 + C3 + Z'
    regression:
        -Type of regression model to fit. Default is log binomial.
         Options include:
            'log':      Log-binomial model. Estimates the Relative Risk (RR)
            'logit':    Logistic (logit) model. Estimates the Odds Ratio (OR). Note, this is only valid when the
                        OR approximates the RR
    ci:
        -Type of confidence interval to return. Default is the delta method. Options include:
            'delta':      Delta method as described by Hosmer and Lemeshow (1992)
            'bootstrap':  bootstrap method (Assmann et al. 1996). The delta method is more time efficient than bootstrap
    b_sample:
        -Number of times to resample to generate bootstrap confidence intervals. Only important if bootstrap confidence
         intervals are requested. Default is 1000
    alpha:
        -Alpha level for confidence interval. Default is 0.05
    decimal:
        -Number of decimal places to display in result. Default is 3

    Example)
    >>>zepid.ICR(df=data,exposure='X',outcome='D',modifier='Z',adjust='var1 + var2')
    """
    df.loc[((df[exposure] == 1) & (df[modifier] == 0)), 'E1M0'] = 1
    df.loc[((df[exposure] != 1) | (df[modifier] != 0)), 'E1M0'] = 0
    df.loc[((df[exposure].isnull()) | (df[modifier].isnull())), 'E1M0'] = 0
    df.loc[((df[exposure] == 0) & (df[modifier] == 1)), 'E0M1'] = 1
    df.loc[((df[exposure] != 0) | (df[modifier] != 1)), 'E0M1'] = 0
    df.loc[((df[exposure].isnull()) | (df[modifier].isnull())), 'E0M1'] = 0
    df.loc[((df[exposure] == 1) & (df[modifier] == 1)), 'E1M1'] = 1
    df.loc[((df[exposure] != 1) | (df[modifier] != 1)), 'E1M1'] = 0
    df.loc[((df[exposure].isnull()) | (df[modifier].isnull())), 'E1M1'] = np.nan
    if regression == 'logit':
        f = sm.families.family.Binomial(sm.families.links.logit)
        print('Note: Using the Odds Ratio to calculate the ICR is only valid when\nthe OR approximates the RR')
    elif regression == 'log':
        f = sm.families.family.Binomial(sm.families.links.log)
    if adjust == None:
        eq = outcome + ' ~ E1M0 + E0M1 + E1M1'
    else:
        eq = outcome + ' ~ E1M0 + E0M1 + E1M1 + ' + adjust
    model = smf.glm(eq, df, family=f).fit()
    print(model.summary())
    em10 = math.exp(model.params['E1M0'])
    em01 = math.exp(model.params['E0M1'])
    em11 = math.exp(model.params['E1M1'])
    em_expect = em10 + em01 - 1
    icr = em11 - em_expect
    zalpha = norm.ppf((1 - alpha / 2), loc=0, scale=1)
    if ci == 'delta':
        cov_matrix = model.cov_params()
        vb10 = cov_matrix.loc['E1M0']['E1M0']
        vb01 = cov_matrix.loc['E0M1']['E0M1']
        vb11 = cov_matrix.loc['E1M1']['E1M1']
        cvb10_01 = cov_matrix.loc['E1M0']['E0M1']
        cvb10_11 = cov_matrix.loc['E1M0']['E1M1']
        cvb01_11 = cov_matrix.loc['E0M1']['E1M1']
        varICR = (((em10 ** 2) * vb10) + ((em01 ** 2) * vb01) + ((em11 ** 2) * vb11) + (
        (em10 * em01 * 2 * cvb10_01)) + (-1 * em10 * em11 * 2 * cvb10_11) + (-1 * em01 * em11 * 2 * cvb01_11))
        icr_lcl = icr - zalpha * math.sqrt(varICR)
        icr_ucl = icr + zalpha * math.sqrt(varICR)
    elif ci == 'bootstrap':
        print('Running bootstrap... please wait...')
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            bse_icr = []
            ul = 1 - alpha / 2
            ll = 0 + alpha / 2
            for i in range(b_sample):
                dfs = df.sample(n=df.shape[0], replace=True)
                try:
                    bmodel = smf.glm(eq, dfs, family=f).fit()
                    em_bexpect = math.exp(bmodel.params['E1M0']) + math.exp(bmodel.params['E0M1']) - 1
                    bicr = math.exp(bmodel.params['E1M1']) - em_bexpect
                    sigma = bicr - icr
                    bse_icr.append(sigma)
                except:
                    bse_icr.append(np.nan)
            bsdf = pd.DataFrame()
            bsdf['sigma'] = bse_icr
            lsig, usig = bsdf['sigma'].dropna().quantile(q=[ll, ul])
            icr_lcl = lsig + icr
            icr_ucl = usig + icr
    else:
        raise ValueError('Please specify a supported confidence interval type')
    print('\n----------------------------------------------------------------------')
    if regression == 'logit':
        print('ICR based on Odds Ratio\t\tAlpha = ' + str(alpha))
        print('Note: Using the Odds Ratio to calculate the ICR is only valid when\nthe OR approximates the RR')
    elif regression == 'log':
        print('ICR based on Risk Ratio\t\tAlpha = ' + str(alpha))
    print('\nICR:\t\t' + str(round(icr, decimal)))
    print('CI:\t\t(' + str(round(icr_lcl, decimal)) + ', ' + str(round(icr_ucl, decimal)) + ')')
    print('----------------------------------------------------------------------')


def Sensitivity(df, test, disease, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """Generates the sensitivity and (1-alpha)% confidence interval, comparing test results to disease status
    from pandas dataframe

    WARNING: Disease & Test must be coded as (1: yes, 0:no)

    test:
        -column name of test results to detect the outcome. Needs to be coded as binary (0,1), where 1 indicates a
        positive test for the individual
    disease:
        -column name of true outcomes status. Needs to be coded as binary (0,1), where 1 indicates the individual
         has the outcome
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the calculated sensitivity. Default is False.

    Example)
    >zepid.Sensitivity(df=data,test='test_disease',disease='true_disease')
    """
    a = df.loc[(df[test] == 1) & (df[disease] == 1)].shape[0]
    b = df.loc[(df[test] == 1) & (df[disease] == 0)].shape[0]
    if print_result is True:
        print(tabulate([["T+", a, b]], headers=['', 'D+', 'D-'], tablefmt='grid'))
        print('----------------------------------------------------------------------')
        sensitivity(a, a + b, alpha=alpha, decimal=decimal, confint='wald', print_result=True)
        print('----------------------------------------------------------------------')
    if return_result is True:
        se = sensitivity(a, a + b, print_result=False, return_result=True)
        return se


def Specificity(df, test, disease, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """Generates the Specificity and (1-alpha)% confidence interval, comparing test results to disease status
    from pandas dataframe

    WARNING: Disease & Test must be coded as (1: yes, 0:no)

    test:
        -column name of test results to detect the outcome. Needs to be coded as binary (0,1), where 1 indicates a
        positive test for the individual
    disease:
        -column name of true outcomes status. Needs to be coded as binary (0,1), where 1 indicates the individual
         has the outcome
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the calculated specificity. Default is False.

    Example)
    >zepid.Specificity(df=data,test='test_disease',disease='true_disease')
    """
    c = df.loc[(df[test] == 0) & (df[disease] == 1)].shape[0]
    d = df.loc[(df[test] == 0) & (df[disease] == 0)].shape[0]
    if print_result:
        print(tabulate([["T-", c, d]], headers=['', 'D+', 'D-'], tablefmt='grid'))
        print('----------------------------------------------------------------------')
        specificity(c, c + d, alpha=alpha, decimal=decimal, confint='wald', print_result=True)
        print('----------------------------------------------------------------------')
    if return_result:
        sp = specificity(c, c + d, print_result=False, return_result=True)
        return sp


def Diagnostics(df, test, disease, alpha=0.05, decimal=3, print_result=True, return_result=False):
    """
    Generates the Sensitivity, Specificity and corresponding (1-alpha)% confidence intervals, comparing test results
    to disease status from pandas dataframe

    WARNING: Disease & Test must be coded as (1: yes, 0:no)

    test:
        -column name of test results to detect the outcome. Needs to be coded as binary (0,1), where 1 indicates a
        positive test for the individual
    disease:
        -column name of true outcomes status. Needs to be coded as binary (0,1), where 1 indicates the individual
         has the outcome
    alpha:
        -Alpha value to calculate two-sided Wald confidence intervals. Default is 95% onfidence interval
    decimal:
        -amount of decimal points to display. Default is 3
    print_result:
        -Whether to print the results. Default is True
    return_result:
        -Whether to return the calculates sensitivity and specificity. Default is False.
    """
    a = df.loc[(df[test] == 1) & (df[disease] == 1)].shape[0]
    b = df.loc[(df[test] == 1) & (df[disease] == 0)].shape[0]
    c = df.loc[(df[test] == 0) & (df[disease] == 1)].shape[0]
    d = df.loc[(df[test] == 0) & (df[disease] == 0)].shape[0]
    if print_result:
        print(tabulate([["T+", a, b], ["T-", c, d]], headers=['', 'D+', 'D-'], tablefmt='grid'))
        print('----------------------------------------------------------------------')
        sensitivity(a, a + b, alpha=alpha, decimal=decimal, confint='wald', print_result=True)
        specificity(c, c + d, alpha=alpha, decimal=decimal, confint='wald', print_result=True)
        print('----------------------------------------------------------------------')
    if return_result:
        se = sensitivity(a, a + b, print_result=False, return_result=True)
        sp = specificity(c, c + d, print_result=False, return_result=True)
        return se, sp


def spline(df, var, n_knots=3, knots=None, term=1, restricted=False):
    """Creates spline dummy variables based on either user specified knot locations or automatically
    determines knot locations based on percentiles. Options are available to set the number of knots,
    location of knots (value), term (linear, quadratic, etc.), and restricted/unrestricted.

    Returns a pandas dataframe containing the spline variables (labeled 0 to n_knots)

    df:
        -pandas dataframe containing the variables of interest
    var:
        -continuous variable to generate spline for
    n_knots:
        -number of knots requested. Options for knots include any positive integer if the location of
         knots are specified. If knot locations are not specified, n_knots must be an integer between
         1 to 7, including both. Default is set to 3
    knots:
        -Location of specified knots in a list. To specify the location of knots, put desired numbers for
         knots into a list. Be sure that the length of the list is the same as the specified number of knots.
         Default is None, so that the function will automatically determine knot locations without user specification
    term:
        -High order term for the spline terms. To calculate a quadratic spline change to 2, cubic spline
         change to 3, etc. Default is 1, so a linear spline is returned
    restricted:
        -Whether to return a restricted spline. Note that the restricted spline returns one less column
         than the number of knots. An unrestricted spline returns the same number of columns as the number of knots.
         Default is False, providing an unrestricted spline

    Example)
    >>>zepid.spline(df=data,var='var1',n_knots=4,term=2,restricted=True)
           rspline0     rspline1   rspline2
    0   9839.409066  1234.154601   2.785600
    1    446.391437     0.000000   0.000000
    2   7107.550298   409.780251   0.000000
    3   4465.272901     7.614501   0.000000
    4  10972.041543  1655.208555  52.167821
    ..          ...          ...        ...
    """
    if knots == None:
        if n_knots == 1:
            knots = [0.5]
        elif n_knots == 2:
            knots = [1 / 3, 2 / 3]
        elif n_knots == 3:
            knots = [0.05, 0.5, 0.95]
        elif n_knots == 4:
            knots = [0.05, 0.35, 0.65, 0.95]
        elif n_knots == 5:
            knots = [0.05, 0.275, 0.50, 0.725, 0.95]
        elif n_knots == 6:
            knots = [0.05, 0.23, 0.41, 0.59, 0.77, 0.95]
        elif n_knots == 7:
            knots = [0.025, 11 / 60, 26 / 75, 0.50, 79 / 120, 49 / 60, 0.975]
        else:
            raise ValueError(
                'When the knot locations are not pre-specified, the number of specified knots must be'
                ' an integer between 1 and 7')
        pts = list(df[var].quantile(q=knots))
    else:
        if n_knots != len(knots):
            raise ValueError('The number of knots and the number of specified knots must match')
        else:
            pass
        pts = knots
    if sorted(pts) != pts:
        raise ValueError('Knots must be in ascending order')
    colnames = []
    sf = df.copy()
    for i in range(len(pts)):
        colnames.append('spline' + str(i))
        sf['spline' + str(i)] = np.where(sf[var] > pts[i], (sf[var] - pts[i]) ** term, 0)
        sf['spline' + str(i)] = np.where(sf[var].isnull(), np.nan, sf['spline' + str(i)])
    if restricted is False:
        return sf[colnames]
    elif restricted is True:
        rsf = sf.copy()
        colnames = []
        for i in range(len(pts) - 1):
            colnames.append('rspline' + str(i))
            rsf['rspline' + str(i)] = np.where(rsf[var] > pts[i],
                                               rsf['spline' + str(i)] - rsf['spline' + str(len(pts) - 1)], 0)
            rsf['rspline' + str(i)] = np.where(rsf[var].isnull(), np.nan, rsf['rspline' + str(i)])
        return rsf[colnames]
    else:
        raise ValueError('restricted must be set to either True or False')


def Table1(df, cols, variable_type, continuous_measure='median', strat_by=None, decimal=3):
    """Code to automatically generate a descriptive table of your study population (often referred to as a
    Table 1). Personally, I hate copying SAS/R/Python output from the interpreter to an Excel or other
    spreadsheet software. This code will generate a pandas dataframe object. This object will be a formatted
    table which can be exported as a CSV, opened in Excel, then final formatting changes/renaming can be done.
    Variables with np.nan values are counted as missing

    Categorical variables will be divided into the unique numbers and have a percent calculated. Additionally,
    missing data will be counted (but is not included in the percent). Additionally, a single categorical variable
    can be used to present the results

    Continuous variables either have median/IQR or mean/SE calculated depending on what is requested. Missing are
    counted as a separate category

    Returns a pandas dataframe object containing a formatted Table 1. It is not recommended that this table is used
    in any part of later analysis, since is id difficult to parse through the table. This function is only meant to
    reduce the amount of copying from output needed.

    df:
        -pandas dataframe object containing all variables of interest
    cols:
        -list of columns of variable names to include in the table. Ex) ['X',var1','var2']
    variable_types:
        -list of strings indicating the variable types. Ex) ['category','continuous','continuous']
         Options
            'category'      :   variable with categories only
            'continuous'    :   continuous variable
    continuous_measure:
        -Whether to use the medians or the means. Default is median
         Options
            'median'    :   returns medians and IQR for continuous variables
            'mean'      :   returns means and SE for continuous variables
    strat_by:
        -What categorical variable to stratify by. Default is None (no stratification)
    decimal:
        -Number of decimals to display in the table. Default is 3

    Example)
    >>>var_types = ['category','category','continuous','continuous','continuous']
    >>>zepid.Table1(df=data,cols=['X','Z','var1','var2','var3'],variable_type=var_types,strat_by='D')
    _                                D=0                             D=1
    __                           % / IQR           n             % / IQR          n

    Variable
    TOTAL                                 310.000000                      74.000000
    X        1.0                0.608187  104.000000            0.692308  27.000000
             0.0                0.391813   67.000000            0.307692  12.000000
             Missing                      139.000000                      35.000000
    Z        1.0                0.722581  224.000000            0.635135  47.000000
             0.0                0.277419   86.000000            0.364865  27.000000
             Missing                        0.000000                       0.000000
    var1              [468.231, 525.312]  497.262978  [481.959, 538.964] 507.286133
             Missing                        0.000000                       0.000000
    var2                [24.454, 25.731]   25.058982      [24.1, 25.607]  24.816898
             Missing                        0.000000                       0.000000
    var3                [24.446, 25.685]   25.037731    [24.388, 25.563]  24.920583
             Missing                        0.000000
    >>>_.to_csv('path/filename.csv')
    """
    # Unstratificed Table 1
    if strat_by == None:
        rlist = []
        for i in cols:
            vn = cols.index(i)
            if continuous_measure == 'median':
                if variable_type[vn] == 'continuous':
                    rf = pd.DataFrame({'n / Median': [np.median(df[i].dropna()), df[i].isnull().sum()],
                                       '% / IQR': [np.percentile(df[i].dropna(), [25, 75]).round(decimals=decimal),
                                                   '']}, index=['', 'Missing'])
                if variable_type[vn] == 'category':
                    x = df[i].value_counts()
                    m = df[i].isnull().sum()
                    rf = pd.DataFrame({'n / Median': x, '% / IQR': x / x.sum()})
                    rf = rf.append(pd.DataFrame({'n / Median': m, '% / IQR': ''}, index=['Missing']))
            elif continuous_measure == 'mean':
                if variable_type[vn] == 'continuous':
                    rf = pd.DataFrame({'n / Mean': [np.mean(df[i].dropna()), df[i].isnull().sum()],
                                       '% / SE': [np.std(df[i].dropna()).round(decimals=decimal), '']},
                                      index=['', 'Missing'])
                if variable_type[vn] == 'category':
                    x = df[i].value_counts(dropna=False)
                    y = df[i].value_counts(dropna=False)
                    rf = pd.DataFrame({'n / Mean': x, '% / SE': y / y.sum()})
            else:
                raise ValueError('median or mean must be specified')
            rlist.append(rf)
        srf = pd.concat(rlist, keys=cols, names=['Variable'])
        if continuous_measure == 'median':
            return srf[['n / Median', '% / IQR']]
        if continuous_measure == 'mean':
            return srf[['n / Mean', '% / SE']]

    # Stratified Table 1
    if strat_by != None:
        v = df[strat_by].dropna().unique()
        slist = []
        nlist = []
        for j in v:
            sf = df.loc[df[strat_by] == j].copy()
            rlist = []
            for i in cols:
                vn = cols.index(i)
                if continuous_measure == 'median':
                    if variable_type[vn] == 'continuous':
                        rf = pd.DataFrame({'n / Median': [np.median(sf[i].dropna()), sf[i].isnull().sum()],
                                           '% / IQR': [np.percentile(sf[i].dropna(), [25, 75]).round(decimals=decimal),
                                                       '']}, index=['', 'Missing'])
                    if variable_type[vn] == 'category':
                        x = sf[i].value_counts()
                        m = sf[i].isnull().sum()
                        rf = pd.DataFrame({'n / Median': x, '% / IQR': x / x.sum()})
                        rf = rf.append(pd.DataFrame({'n / Median': m, '% / IQR': ''}, index=['Missing']))
                if continuous_measure == 'mean':
                    if variable_type[vn] == 'continuous':
                        rf = pd.DataFrame({'n / Mean': [np.mean(sf[i].dropna()), sf[i].isnull().sum()],
                                           '% / SE': [np.std(sf[i].dropna()).round(decimals=decimal), '']},
                                          index=['', 'Missing'])
                    if variable_type[vn] == 'category':
                        x = sf[i].value_counts()
                        m = sf[i].isnull().sum()
                        rf = pd.DataFrame({'n / Mean': x, '% / SD': x / x.sum()})
                        rf = rf.append(pd.DataFrame({'n / Mean': m, '% / SD': ''}, index=['Missing']))
                rlist.append(rf)
            if continuous_measure == 'median':
                c = pd.DataFrame({'n / Median': len(sf), '% / IQR': ''}, index=[''])
            if continuous_measure == 'mean':
                c = pd.DataFrame({'n / Mean': len(sf), '% / SD': ''}, index=[''])
            rff = pd.concat([c] + rlist, keys=['TOTAL'] + cols, names=['Variable'], axis=0)
            slist.append(rff)
            if continuous_measure == 'median':
                nlist.append((strat_by + '=' + str(j), '% / IQR'))
            if continuous_measure == 'mean':
                nlist.append((strat_by + '=' + str(j), '% / SD'))
            nlist.append((strat_by + '=' + str(j), 'n'))
        index = pd.MultiIndex.from_tuples(nlist, names=['_', '__'])
        srf = pd.concat(slist, keys=cols, names=['Variable'], axis=1)
        srf.columns = index
        return srf
